from collections import Counter, defaultdict
from typing import Any, Dict, List, Tuple, NamedTuple, Optional

import glypy

import glycopeptidepy
from glycopeptidepy.structure import sequence, modification, SequencePosition
from glycopeptidepy.structure.modification import GlycanFragment, Modification
from glycopeptidepy.structure.fragment import IonSeries, OxoniumIon, StubFragment, PeptideFragment
from glycopeptidepy.utils import simple_repr
from glycopeptidepy.utils.collectiontools import CountTable

from glypy.plot import plot as draw_tree, SNFGNomenclature
from glypy.plot.common import MonosaccharidePatch

from .colors import darken, cnames, hex2color, get_color

import numpy as np

from matplotlib import (
    patheffects,
    pyplot as plt,
    patches as mpatch,
    textpath,
    font_manager,
    path as mpath,
    transforms as mtransform,
    collections as mcollection,
)


def flatten(x):
    return [xj for xi in x for xj in ([xi] if not isinstance(xi, (tuple, list)) else flatten(xi))]


def centroid(path):
    point = np.zeros_like(path.vertices[0])
    c = 0.0
    for p in path.vertices:
        point += p
        c += 1
    return point / c


# def apply_transform(t, ax, scale=(1, 1), translate=(0, 0)):
#     cent = centroid(t.get_path())
#     trans = mtransform.Affine2D().scale(*scale)
#     new_cent = centroid(t.get_path().transformed(trans))
#     shift = cent - new_cent
#     trans.translate(*(shift + translate))
#     trans += ax.transData
#     t.set_transform(trans)
def apply_transform(t: mpatch.PathPatch, ax, scale=(1, 1), translate=(0, 0), rotate_deg=None):
    cent = centroid(t.get_path())
    trans = mtransform.Affine2D()
    if rotate_deg is not None:
        trans.translate(*(-cent))
        trans.rotate_deg(rotate_deg)
        trans.translate(*(cent))
    trans.scale(*scale)
    new_cent = centroid(t.get_path().transformed(trans))
    shift = cent - new_cent
    trans.translate(*(shift + translate))
    t.set_path(t.get_path().transformed(trans))


font_options = font_manager.FontProperties(family="monospace")
glycan_symbol_grammar = SNFGNomenclature()


class BBox(NamedTuple):
    xmin: float
    ymin: float
    xmax: float
    ymax: float

    def expand(self, xmin, ymin, xmax, ymax) -> "BBox":
        return self.__class__(
            min(xmin, self.xmin),
            min(ymin, self.ymin),
            max(xmax, self.xmax),
            max(ymax, self.ymax),
        )

    @classmethod
    def from_axes(cls, ax: plt.Axes):
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        return BBox(xlim[0], ylim[0], xlim[1], ylim[1])

    @classmethod
    def from_path(cls, path: mpath.Path):
        return bbox_path(path)

    def relative_to_point(self, x, y):
        return BBox(self.xmin - x, self.ymin - y, self.xmax - x, self.ymax - y)

    def as_array(self):
        return np.array(self).reshape((-1, 2))


def bbox_path(path):
    nodes = path.vertices
    if nodes.size == 0:
        return BBox(0, 0, 0, 0)
    xmin = nodes[:, 0].min()
    xmax = nodes[:, 0].max()
    ymin = nodes[:, 1].min()
    ymax = nodes[:, 1].max()
    return BBox(xmin, ymin, xmax, ymax)


class SequencePositionGlyph(object):
    position: SequencePosition
    index: int
    _ax: plt.Axes
    x: float
    y: float
    options: Dict[str, Any]

    def __init__(self, position, index, x, y, patch=None, ax=None, **kwargs):
        self.position = position
        self.index = index
        self.x = x
        self.y = y
        self._patch = patch
        self._ax = ax
        self.options = kwargs

    __repr__ = simple_repr

    def render(self, ax=None):
        if ax is None:
            ax = self._ax
        else:
            self._ax = ax

        symbol = self.position.amino_acid.symbol

        color = self.options.get("color", "black")
        if self.position.modifications:
            color = darken(get_color(self.position.modifications[0]))
            label = self.position.modifications[0].name
        else:
            label = None

        tpath = textpath.TextPath((self.x, self.y), symbol, size=self.options.get("size"), prop=font_options)
        tpatch = mpatch.PathPatch(tpath, color=color, lw=self.options.get("lw", 0), label=label)
        self._patch = tpatch
        ax.add_patch(tpatch)
        return ax

    def set_transform(self, transform):
        self._patch.set_transform(transform)

    def centroid(self):
        path = self._patch.get_path()
        return centroid(path)

    def bbox(self):
        return bbox_path(self._patch.get_path())


class GlycanCompositionGlyphs(object):
    glycan_composition: glypy.GlycanComposition
    ax: plt.Axes
    x: float
    xend: float
    y: float
    spacing: float
    options: Dict[str, Any]

    patches: List[MonosaccharidePatch]
    layers: List[List[mpatch.PathPatch]]

    def __init__(self, glycan_composition, x, y, ax, vertical=False, spacing: float = 1.0, **kwargs):
        self.glycan_composition = glycan_composition
        self.ax = ax
        self.x = x
        self.xend = x
        self.y = y
        self.yend = y
        self.vertical = vertical
        self.options = kwargs
        self.patches = []
        self.layers = []
        self.spacing = spacing

    def render(self):
        x = self.x
        y = self.y
        glyphs = []
        layers = []
        for mono, count in self.glycan_composition.items():
            layer = []
            glyph = glycan_symbol_grammar.draw(mono, x, y, ax=self.ax, scale=(0.4, 0.4))
            x += 0.6 * self.spacing
            layer.extend(flatten(glyph.shape_patches))
            glyph = glycan_symbol_grammar.draw_text(
                self.ax, x, y - 0.17, r"$\times %d$" % count, center=False, fontsize=22, zorder=100, lw=0
            )
            if isinstance(glyph, tuple):
                for g in glyph:
                    g.set_lw(0)
            else:
                glyph.set_lw(0)
            if self.vertical:
                y += 1.0 * self.spacing
                x -= 0.6 * self.spacing
            else:
                x += 1.75
            layer.extend(glyph)
            layers.append(layer)
            glyphs.extend(layer)
        if self.vertical:
            x += 0.95 + (0.6 * self.spacing)
        self.xend = x
        self.yend = y
        self.patches = glyphs
        self.layers = layers

    def set_transform(self, transform):
        for patch in self.patches:
            patch.set_transform(transform)

    def bbox(self):
        if not self.patches:
            return self.x, self.y, self.xend, self.y

        flat_patch_iter = (
            p for patch in self.patches for p in ([patch] if not isinstance(patch, (tuple, list)) else patch)
        )
        first = next(flat_patch_iter)
        bbox = bbox_path(first.get_path())
        for patch in flat_patch_iter:
            bbox = bbox.expand(*bbox_path(patch.get_path()))
        return bbox


class GlycanCompositionGlyphsOverlaid(GlycanCompositionGlyphs):
    def render(self):
        x = self.x
        y = self.y
        glyphs = []
        layers = []
        for mono, count in self.glycan_composition.items():
            layer = []
            glyph = glycan_symbol_grammar.draw(mono, x, y, ax=self.ax, scale=(0.4, 0.4))
            # x += 0.6 * self.spacing
            layer.extend(flatten(glyph.shape_patches))
            glyph = glycan_symbol_grammar.draw_text(
                self.ax, x - 0.15, y - 0.17, r"$%d$" % count, center=False, fontsize=22, zorder=100, lw=0
            )
            if isinstance(glyph, tuple):
                for g in glyph:
                    g.set_lw(0)
            else:
                glyph.set_lw(0)
            if self.vertical:
                y += 1.0 * self.spacing
                # x -= 0.6 * self.spacing
            else:
                x += 1.75
            layer.extend(glyph)
            layers.append(layer)
            glyphs.extend(layer)
        if self.vertical:
            x += 0.95 + (0.6 * self.spacing)
        self.xend = x
        self.yend = y
        self.patches = glyphs
        self.layers = layers


class MonosaccharideSequence(object):
    monosaccharides: List[glypy.MonosaccharideResidue]
    ax: plt.Axes
    x: float
    xend: float
    y: float
    spacing: float
    options: Dict[str, Any]

    patches: List[MonosaccharidePatch]
    layers: List[List[mpatch.PathPatch]]

    def __init__(self, monosaccharides, x, y, ax, spacing: float = 1.0, **kwargs):
        self.monosaccharides = monosaccharides
        self.ax = ax
        self.x = x
        self.xend = x
        self.y = y
        self.yend = y
        self.options = kwargs
        self.patches = []
        self.layers = []
        self.spacing = spacing

    def render(self):
        x = self.x
        y = self.y
        layers = []
        for mono in self.monosaccharides:
            layer = []
            glyph = glycan_symbol_grammar.draw(mono, x, y, ax=self.ax, scale=(0.4, 0.4), zorder=100)
            y += 1.0 * self.spacing
            layer.extend(flatten(glyph.shape_patches))
            layers.append(layer)
        self.layers = layers
        self.patches = [o for layer in layers for o in layer]

    def set_transform(self, transform):
        for patch in self.patches:
            patch.set_transform(transform)

    def bbox(self):
        if not self.patches:
            return self.x, self.y, self.xend, self.y

        flat_patch_iter = (
            p for patch in self.patches for p in ([patch] if not isinstance(patch, (tuple, list)) else patch)
        )
        first = next(flat_patch_iter)
        bbox = bbox_path(first.get_path())
        for patch in flat_patch_iter:
            bbox = bbox.expand(*bbox_path(patch.get_path()))
        return bbox


class IonAnnotationGlyphBase:
    xaspect: float = 32
    yaspect: float = 24

    x: float
    y: float
    ax: plt.Axes

    xscale: float
    yscale: float
    charge: int

    peptide_patch: mpatch.PathPatch
    charge_patch: mpatch.PathPatch
    patches: List[mpatch.PathPatch]
    glycan_patch_layers: List[List[mpatch.PathPatch]]

    size: int

    base_size: int = 22

    def __init__(self, x: float, y: float, ax: plt.Axes, xscale: float, yscale: float, charge: int, size: int = 22):
        self.x = x
        self.y = y
        self.ax = ax
        self.charge = charge
        self.xscale = xscale
        self.yscale = yscale
        self.patches = []
        self.glycan_patch_layers = []
        self.peptide_patch = None
        self.charge_patch = None
        self.size = size

    def __repr__(self):
        return f"{self.__class__.__name__}({self.x}, {self.y}, {self.fragment.name})"

    def draw_charge_patch(self):
        y = 0
        for patch in [self.peptide_patch] + self.patches:
            y = max(bbox_path(patch.get_path()).ymax, y)
        if abs(self.charge) == 1:
            return
        if self.charge > 0:
            label = f"+{self.charge}"
        else:
            label = str(self.charge)

        (glyph,) = glycan_symbol_grammar.draw_text(
            self.ax, -0.45, y + 0.15, label, center=False, fontsize=16, zorder=100, lw=0
        )
        p = glyph.get_path()
        center = centroid(p)
        p = (
            p.transformed(mtransform.Affine2D().translate(*-center))
            .transformed(mtransform.Affine2D().scale((self.size + 4) / self.base_size))
            .transformed(mtransform.Affine2D().translate(*center))
        )
        lower = bbox_path(p).ymin
        upshift = y - lower
        if upshift > 0:
            upshift *= 2
            p = p.transformed(mtransform.Affine2D().translate(0, upshift))
        glyph.set_path(p)
        self.charge_patch = glyph

    def get_glycan_composition(self):
        frag = self.fragment

        if frag.glycosylation:
            return frag.glycosylation
        elif hasattr(frag, "modification_dict"):
            mods = CountTable()
            for mod, v in frag.modification_dict.items():
                if isinstance(mod, GlycanFragment):
                    mods += mod.fragment.glycan_composition
                elif (
                    isinstance(mod, Modification)
                    and glycopeptidepy.ModificationCategory.glycosylation in mod.rule.categories
                ) or glycopeptidepy.ModificationCategory.glycosylation in mod.categories:
                    mods[glypy.MonosaccharideResidue.from_iupac_lite(mod.name)] += v
            mods = glypy.GlycanComposition(mods)
            return mods
        else:
            return glypy.GlycanComposition()

    def render(self):
        raise NotImplementedError()

    def bbox(self):
        patches = self.get_all_patches()

        flat_patch_iter = iter(patches)
        first = next(flat_patch_iter)
        bbox = bbox_path(first.get_path())
        for patch in flat_patch_iter:
            bbox = bbox.expand(*bbox_path(patch.get_path()))
        return bbox

    def get_all_patches(self) -> List[mpatch.PathPatch]:
        patches = list(self.patches)
        if self.peptide_patch:
            patches = [self.peptide_patch] + patches
        if self.charge_patch:
            patches.append(self.charge_patch)
        return patches

    def translate(self, x, y):
        self.x += x
        self.y += y
        for patch in self.get_all_patches():
            patch.set_path(patch.get_path().transformed(mtransform.Affine2D().translate(x, y)))


class GlycopeptideStubFragmentGlyph(IonAnnotationGlyphBase):
    fragment: StubFragment

    def __init__(
        self,
        x: float,
        y: float,
        ax: plt.Axes,
        fragment: StubFragment,
        xscale: float,
        yscale: float,
        charge: int,
        size: int = 22,
    ):
        super().__init__(x, y, ax, xscale, yscale, charge, size)
        self.fragment = fragment

    def draw_peptide_text(self):
        x = 0
        y = 0
        if self.get_glycan_composition():
            label = "pep+"
        else:
            label = "pep"
        glyph = glycan_symbol_grammar.draw_text(
            self.ax, x, y, label, center=False, fontsize=self.base_size, zorder=100, lw=0
        )
        glyph = glyph[0]
        path = glyph.get_path()
        glyph.set_path(
            path.transformed(mtransform.Affine2D().scale(self.size / self.base_size))
            .transformed(mtransform.Affine2D().rotate_deg(90))
            .transformed(mtransform.Affine2D().translate(0, 0))
        )
        self.peptide_patch = glyph

    def draw_glycan_composition(self):
        y = bbox_path(self.peptide_patch.get_path()).ymax
        art = GlycanCompositionGlyphsOverlaid(
            self.get_glycan_composition(),
            0 - 0.30,
            y + (0.5 * self.size / self.base_size),
            self.ax,
            spacing=self.size / self.base_size,
            vertical=True,
        )
        art.render()
        ytop = y
        for layer in art.layers:
            ybottom = None
            for patch in layer:
                p: mpath.Path = patch.get_path()
                center = centroid(p)
                p: mpath.Path = (
                    p.transformed(mtransform.Affine2D().translate(*-center))
                    .transformed(mtransform.Affine2D().scale(self.size / self.base_size))
                    .transformed(mtransform.Affine2D().translate(*center))
                    .transformed(mtransform.Affine2D().scale(0.65, 1.0))
                )
                bb_next = bbox_path(p)
                if ybottom is None:
                    ybottom = bb_next.ymin
                elif ybottom > bb_next.ymin:
                    ybottom = bb_next.ymin
                patch.set_path(p)

            if ybottom <= ytop:
                for patch in layer:
                    p: mpath.Path = patch.get_path()
                    p = p.transformed(mtransform.Affine2D().translate(0, (ytop - ybottom) + 0.01 * self.yscale))
                    patch.set_path(p)
            for patch in layer:
                p: mpath.Path = patch.get_path()
                bb_next = bbox_path(p)
                ytop = max(ytop, bb_next.ymax)
            patch.set_path(p)
        self.patches = art.patches
        self.glycan_patch_layers = art.layers

    def render(self):
        self.draw_peptide_text()
        self.draw_glycan_composition()
        self.draw_charge_patch()

        patches = [self.peptide_patch] + self.patches
        if self.charge_patch:
            patches.append(self.charge_patch)

        for glyph in patches:
            glyph.set_path(
                glyph.get_path()
                .transformed(mtransform.Affine2D().scale(self.xscale / self.xaspect, self.yscale / self.yaspect))
                .transformed(mtransform.Affine2D().translate(self.x + 0.15 * self.xscale / self.xaspect, self.y))
            )
            glyph.set_path_effects(
                [
                    patheffects.Stroke(linewidth=0.5, foreground="white"),
                    patheffects.Normal(),
                ]
            )


class OxoniumIonGlyph(IonAnnotationGlyphBase):
    fragment: OxoniumIon
    loss_patch: mpatch.PathPatch

    def __init__(
        self,
        x: float,
        y: float,
        ax: plt.Axes,
        fragment: OxoniumIon,
        xscale: float,
        yscale: float,
        charge: int,
        size: int = 22,
    ):
        super().__init__(x, y, ax, xscale, yscale, charge, size)
        self.fragment = fragment
        self.loss_patch = None

    def get_glycan_composition(self):
        return glypy.GlycanComposition(Counter(self.fragment.monosaccharides))

    def draw_glycan_composition(self):
        y = 0
        if len(self.fragment.monosaccharides) > 3:
            art = GlycanCompositionGlyphsOverlaid(
                self.get_glycan_composition(),
                -0.2,
                y + (0.5 * self.size / self.base_size),
                self.ax,
                spacing=self.size / self.base_size,
                vertical=True,
            )
        else:
            art = MonosaccharideSequence(
                self.fragment.monosaccharides,
                -0.2,
                y + (0.5 * self.size / self.base_size),
                self.ax,
                spacing=self.size / self.base_size,
                vertical=True,
            )
        art.render()
        ytop = y
        for layer in art.layers:
            ybottom = None
            for patch in layer:
                p: mpath.Path = patch.get_path()
                center = centroid(p)
                p: mpath.Path = (
                    p.transformed(mtransform.Affine2D().translate(*-center))
                    .transformed(mtransform.Affine2D().scale(self.size / self.base_size))
                    .transformed(mtransform.Affine2D().translate(*center))
                    .transformed(mtransform.Affine2D().scale(0.65, 1.0))
                )
                bb_next = bbox_path(p)
                if ybottom is None:
                    ybottom = bb_next.ymin
                elif ybottom > bb_next.ymin:
                    ybottom = bb_next.ymin
                patch.set_path(p)

            if ybottom <= ytop:
                for patch in layer:
                    p: mpath.Path = patch.get_path()
                    p = p.transformed(mtransform.Affine2D().translate(0, (ytop - ybottom) + 0.01 * self.yscale))
                    patch.set_path(p)
            for patch in layer:
                p: mpath.Path = patch.get_path()
                bb_next = bbox_path(p)
                ytop = max(ytop, bb_next.ymax)
            patch.set_path(p)
        self.patches = art.patches
        self.glycan_patch_layers = art.layers

    def draw_loss(self):
        y = 0
        xs = []
        for patch in self.patches:
            y = max(bbox_path(patch.get_path()).ymax, y)
            x = centroid(patch.get_path())[0]
            xs.append(x)
        if xs:
            x = np.mean(x)
        else:
            x = 0.0
        if self.fragment.chemical_shift is None:
            return

        label = self.fragment.chemical_shift.name
        glyph = glycan_symbol_grammar.draw_text(
            self.ax, x, y, label, center=False, fontsize=self.base_size, zorder=100, lw=0
        )
        glyph = glyph[0]
        path = glyph.get_path()
        glyph.set_path(
            path.transformed(mtransform.Affine2D().scale(self.size / self.base_size)).transformed(
                mtransform.Affine2D().rotate_deg_around(x, y, 90.0)
            )
        )
        self.loss_patch = glyph

    def render(self):
        self.draw_glycan_composition()
        self.draw_loss()

        patches = self.patches

        for glyph in patches:
            glyph.set_path(
                glyph.get_path()
                .transformed(mtransform.Affine2D().scale(self.xscale / self.xaspect, self.yscale / self.yaspect))
                .transformed(mtransform.Affine2D().translate(self.x + 0.15 * self.xscale / self.xaspect, self.y))
            )
            glyph.set_path_effects(
                [
                    patheffects.Stroke(linewidth=0.5, foreground="white"),
                    patheffects.Normal(),
                ]
            )

        if self.loss_patch is not None:
            xs = []
            for patch in patches:
                x = centroid(patch.get_path())[0]
                xs.append(x)
            if xs:
                x = np.mean(x)
            else:
                x = 0.0

            glyph = self.loss_patch

            # Center x at 0.0
            glyph.set_path(
                glyph.get_path().transformed(mtransform.Affine2D().translate(-centroid(glyph.get_path())[0], 0.0))
            )

            glyph.set_path(
                # Scale paths
                glyph.get_path()
                .transformed(mtransform.Affine2D().scale(self.xscale / self.xaspect, self.yscale / self.yaspect))
                .transformed(
                    # Move to expected position
                    mtransform.Affine2D().translate(x, self.y)
                )
            )

    def get_all_patches(self) -> List[mpatch.PathPatch]:
        patches = list(self.patches)
        if self.loss_patch and self.loss_patch not in patches:
            patches.append(self.loss_patch)
        return patches

    def bbox(self):
        patches = self.get_all_patches()
        flat_patch_iter = iter(patches)
        first = next(flat_patch_iter)
        bbox = bbox_path(first.get_path())
        for patch in flat_patch_iter:
            bbox = bbox.expand(*bbox_path(patch.get_path()))
        return bbox


class PeptideFragmentGlyph(IonAnnotationGlyphBase):
    fragment: PeptideFragment

    def __init__(
        self,
        x: float,
        y: float,
        ax: plt.Axes,
        fragment: glycopeptidepy.PeptideFragment,
        xscale: float,
        yscale: float,
        charge: int,
        size: int = 22,
    ):
        super().__init__(x, y, ax, xscale, yscale, charge, size)
        self.fragment = fragment

    def draw_peptide_text(self):
        x = 0
        y = 0
        label = self.fragment.base_name()
        glyph = glycan_symbol_grammar.draw_text(
            self.ax, x, y, label, center=False, fontsize=self.base_size, zorder=100, lw=0
        )
        glyph = glyph[0]
        path = glyph.get_path()
        bbox = bbox_path(path)
        xc = (bbox.xmin + bbox.xmax) / 2
        yc = (bbox.ymin + bbox.ymax) / 2
        glyph.set_path(
            path.transformed(mtransform.Affine2D().scale(self.size / self.base_size)).transformed(
                mtransform.Affine2D().translate(-xc, yc)
            )
        )
        self.peptide_patch = glyph

    def draw_glycan_composition(self):
        gc = self.get_glycan_composition()
        bb = bbox_path(self.peptide_patch.get_path())
        y = bb.ymax
        art = GlycanCompositionGlyphsOverlaid(
            gc,
            0 - 0.1,
            y + (0.5 * self.size / self.base_size),
            self.ax,
            vertical=True,
            spacing=self.size / self.base_size,
        )
        art.render()
        ytop = y
        for layer in art.layers:
            ybottom = None
            for patch in layer:
                p: mpath.Path = patch.get_path()
                center = centroid(p)
                p: mpath.Path = (
                    p.transformed(mtransform.Affine2D().translate(*-center))
                    .transformed(mtransform.Affine2D().scale(self.size / self.base_size))
                    .transformed(mtransform.Affine2D().translate(*center))
                    .transformed(mtransform.Affine2D().scale(0.65, 1))
                )
                bb_next = bbox_path(p)
                if ybottom is None:
                    ybottom = bb_next.ymin
                elif ybottom > bb_next.ymin:
                    ybottom = bb_next.ymin
                patch.set_path(p)

            if ybottom <= ytop:
                for patch in layer:
                    p: mpath.Path = patch.get_path()
                    p = p.transformed(mtransform.Affine2D().translate(0, ytop - ybottom))
                    patch.set_path(p)
            for patch in layer:
                p: mpath.Path = patch.get_path()
                bb_next = bbox_path(p)
                ytop = max(ytop, bb_next.ymax)
            patch.set_path(p)
        self.patches = art.patches
        self.glycan_path_layers = art.layers

    def draw_charge_patch(self):
        y = 0
        for patch in [self.peptide_patch] + self.patches:
            y = max(bbox_path(patch.get_path()).ymax, y)
        if abs(self.charge) == 1:
            return
        if self.charge > 0:
            label = f"+{self.charge}"
        else:
            label = str(self.charge)

        (glyph,) = glycan_symbol_grammar.draw_text(self.ax, -0.2, y + 0.1, label, center=False, fontsize=16, lw=0)
        p = glyph.get_path()
        center = centroid(p)
        p = (
            p.transformed(mtransform.Affine2D().translate(*-center))
            .transformed(mtransform.Affine2D().scale((self.size + 4) / self.base_size))
            .transformed(mtransform.Affine2D().translate(*center))
        )
        lower = bbox_path(p).ymin
        upshift = y - lower
        if upshift > 0:
            upshift *= 2
            p = p.transformed(mtransform.Affine2D().translate(0, upshift))
        glyph.set_path(p)
        self.charge_patch = glyph

    def render(self):
        self.draw_peptide_text()
        self.draw_glycan_composition()
        self.draw_charge_patch()

        patches = [self.peptide_patch] + self.patches
        if self.charge_patch:
            patches.append(self.charge_patch)

        for i, glyph in enumerate(patches):
            p = glyph.get_path()
            p = p.transformed(
                mtransform.Affine2D().scale(self.xscale / self.xaspect, self.yscale / self.yaspect)
            ).transformed(mtransform.Affine2D().translate(self.x + 0.25 * self.xscale / self.xaspect, self.y))
            if i == 0:
                bbox = bbox_path(p)
                shift = (bbox.xmin - self.x) / 2
                p = p.transformed(mtransform.Affine2D().translate(shift, 0))
            glyph.set_path(p)
            glyph.set_path_effects(
                [
                    patheffects.Stroke(linewidth=0.5, foreground="white"),
                    patheffects.Normal(),
                ]
            )
        cent = centroid(self.peptide_patch.get_path())
        centx = cent[0]
        xoffset = self.x - centx
        for glyph in patches:
            p = glyph.get_path()
            p = p.transformed(mtransform.Affine2D().translate(xoffset, 0))
            glyph.set_path(p)


class FragmentStroke:
    index: int

    linewidth: float
    x: float
    top: float
    bottom: float

    n_length: float
    c_length: float

    v_step: float
    stroke_slope: float

    c_labels: List[str]
    n_labels: List[str]

    main_color: str
    c_colors: List[str]
    n_colors: List[str]

    main_stroke: mpatch.PathPatch
    c_strokes: List[mpatch.PathPatch]
    n_strokes: List[mpatch.PathPatch]

    use_collection: bool = True
    _collection: mcollection.PatchCollection
    ax: plt.Axes
    options: Dict[str, Any]

    def __init__(
        self,
        index: int,
        x: float,
        top: float,
        bottom: float,
        n_labels=None,
        c_labels=None,
        main_color=None,
        n_colors=None,
        c_colors=None,
        ax=None,
        v_step: float = 0.2,
        stroke_slope: float = 0.2,
        n_length: float = 0.8,
        c_length: float = 0.8,
        options: Dict[str, Any] = None,
        **kwargs,
    ):
        if options is None:
            options = {}
        options.update(kwargs)

        self.index = index
        self.x = x
        self.top = top
        self.bottom = bottom

        self.v_step = v_step
        self.stroke_slope = stroke_slope

        self.c_length = c_length
        self.n_length = n_length

        self.c_labels = c_labels or []
        self.n_labels = n_labels or []
        self.main_color = main_color or "black"
        self.c_colors = c_colors or ["black"] * len(self.c_labels)
        self.n_colors = n_colors or ["black"] * len(self.n_labels)

        self.main_stroke = None
        self.c_strokes = []
        self.n_strokes = []

        self.ax = ax
        self.options = options
        self.options.setdefault("lw", 2)
        self._collection = None

    def has_annotations(self):
        return self.has_n_annotations() or self.has_c_annotations()

    def has_n_annotations(self):
        return bool(self.n_labels)

    def has_c_annotations(self):
        return bool(self.c_labels)

    def add_n_label(self, label: str, color="black"):
        self.n_labels.append(label)
        self.n_colors.append(color)

    def add_c_label(self, label: str, color="black"):
        self.c_labels.append(label)
        self.c_colors.append(color)

    def draw_main_stroke(self):
        p = mpath.Path(
            [[self.x, self.top], [self.x, self.bottom], [self.x, self.bottom]],
            [mpath.Path.MOVETO, mpath.Path.LINETO, mpath.Path.STOP],
        )

        pp = mpatch.PathPatch(
            p,
            facecolor=self.main_color,
            edgecolor=self.main_color,
            lw=self.options["lw"],
            capstyle="round" if self.stroke_slope != 0 else "projecting",
        )
        if not self.use_collection:
            self.ax.add_patch(pp)
        self.main_stroke = pp

    def draw_c_strokes(self):
        v_step = self.v_step
        stroke_slope = self.stroke_slope
        for i, (c_label, color) in enumerate(zip(self.c_labels, self.c_colors), 1):
            if color is None or color == "none":
                continue
            p = mpath.Path(
                [
                    [self.x, self.top + v_step * (i - 1)],
                    [self.x + self.c_length, self.top + v_step * (i - 1) + stroke_slope],
                    [self.x + self.c_length, self.top + v_step * (i - 1) + stroke_slope],
                ],
                [mpath.Path.MOVETO, mpath.Path.LINETO, mpath.Path.STOP],
            )
            pp = mpatch.PathPatch(
                p,
                facecolor=color,
                edgecolor=color,
                lw=self.options["lw"],
                capstyle="round" if self.stroke_slope != 0 else "projecting",
            )
            if not self.use_collection:
                self.ax.add_patch(pp)
            self.c_strokes.append(pp)

    def draw_n_strokes(self):
        v_step = self.v_step
        for i, (n_label, color) in enumerate(zip(self.n_labels, self.n_colors), 1):
            if color is None or color == "none":
                continue
            p = mpath.Path(
                [
                    [self.x, self.bottom - v_step * (i - 1)],
                    [self.x - self.n_length, self.bottom - v_step * (i - 1) - self.stroke_slope],
                    [self.x - self.n_length, self.bottom - v_step * (i - 1) - self.stroke_slope],
                ],
                [mpath.Path.MOVETO, mpath.Path.LINETO, mpath.Path.STOP],
            )
            pp = mpatch.PathPatch(
                p,
                facecolor=color,
                edgecolor=color,
                lw=self.options["lw"],
                capstyle="round" if self.stroke_slope != 0 else "projecting",
            )
            if not self.use_collection:
                self.ax.add_patch(pp)
            self.n_strokes.append(pp)

    def draw(self):
        self.draw_main_stroke()
        self.draw_c_strokes()
        self.draw_n_strokes()
        if self.c_labels:
            self.draw_c_label()
        if self.n_labels:
            self.draw_n_label()
        if self.use_collection:
            collect = self.as_patch_collection()
            self.ax.add_collection(collect)
            self._collection = collect

    def draw_c_label(self):
        if self.c_strokes:
            p = self.c_strokes[-1]
            base = bbox_path(p.get_path()).ymax
            label_y = base + 0.2
        else:
            i = len(self.c_strokes)
            label_y = self.top + (0.2) * i
        skipped = 0
        seen = set()
        for j, text in enumerate(self.c_labels):
            text = str(text)
            if text in seen or not text:
                skipped += 1
                continue
            seen.add(text)
            tpath = textpath.TextPath(
                (self.x, label_y + (j - skipped) * (self.v_step or 0.2)),
                text,
                size=0.45,
                prop=font_options,
            )
            # bbox = bbox_path(tpath)
            # dy = (bbox.ymax - bbox.ymin) / 2 + 0.2 * (j - skipped)
            dy = (0.2) * (j - skipped)
            tpath = tpath.transformed(mtransform.Affine2D().translate(0, dy))
            tpatch = mpatch.PathPatch(tpath, color="black", lw=0)
            if not self.use_collection:
                self.ax.add_patch(tpatch)
            self.c_strokes.append(tpatch)

    def draw_n_label(self):
        if self.n_strokes:
            p = self.n_strokes[-1]
            base = bbox_path(p.get_path()).ymin
            label_y = base - 0.1
        else:
            i = len(self.n_strokes)
            label_y = self.bottom - (self.v_step or 0.2) * i
        skipped = 0
        seen = set()
        for j, text in enumerate(self.n_labels):
            text = str(text)
            if text in seen or not text:
                skipped += 1
                continue
            seen.add(text)
            tpath = textpath.TextPath(
                (self.x - self.c_length, label_y - (j - skipped) * (self.v_step or 0.2)),
                text,
                size=0.45,
                prop=font_options,
            )
            bbox = bbox_path(tpath)
            dy = (bbox.ymax - bbox.ymin) + 0.16 * (j - skipped)
            tpath = tpath.transformed(mtransform.Affine2D().translate(0, -dy))

            tpatch = mpatch.PathPatch(tpath, color="black", lw=0)
            if not self.use_collection:
                self.ax.add_patch(tpatch)
            self.n_strokes.append(tpatch)

    def bbox(self):
        bbox = bbox_path(self.main_stroke.get_path())
        for stroke in self.n_strokes + self.c_strokes:
            bbox = bbox.expand(*bbox_path(stroke.get_path()))
        return bbox

    def as_patch_collection(self):
        strokes = [self.main_stroke] + self.n_strokes + self.c_strokes
        collect = mcollection.PatchCollection(
            strokes, match_original=True, capstyle="round" if self.stroke_slope != 0 else "projecting"
        )
        return collect

    def set_transform(self, transform: mtransform.Affine2D):
        if self.main_stroke:
            self.main_stroke.set_transform(transform)
        for stroke in self.c_strokes + self.n_strokes:
            stroke.set_transform(transform)
        if self._collection is not None:
            self._collection.set_transform(transform)


class SequenceGlyph(object):
    sequence: glycopeptidepy.PeptideSequence
    sequence_position_glyphs: List[SequencePositionGlyph]
    glycan_composition_glyphs: Optional[GlycanCompositionGlyphs]

    step_coefficient: float
    x: float
    y: float
    size: float
    draw_glycan: bool

    ax: plt.Axes

    # Old and possibly unused
    multi_tier_annotation: bool
    label_fragments: bool

    def __init__(
        self, peptide, ax=None, size=1, step_coefficient=1.0, draw_glycan=False, label_fragments=True, **kwargs
    ):
        if not isinstance(peptide, sequence.PeptideSequenceBase):
            peptide = sequence.PeptideSequence(peptide)
        self.sequence = peptide
        self._fragment_index = None
        self.ax = ax
        self.size = size
        self.step_coefficient = step_coefficient
        self.sequence_position_glyphs = []
        self.annotations = []
        self.x = kwargs.get("x", 1)
        self.y = kwargs.get("y", 1)
        self.options = kwargs
        self.next_at_height = dict(c_term=defaultdict(float), n_term=defaultdict(float))
        self.label_fragments = label_fragments
        self.multi_tier_annotation = False
        self.draw_glycan = draw_glycan
        self.glycan_composition_glyphs = None
        self.render()

    def transform(self, transform):
        for glyph in self.sequence_position_glyphs:
            glyph.set_transform(transform + self.ax.transData)
        for annot in self.annotations:
            annot.set_transform(transform + self.ax.transData)
        if self.draw_glycan:
            self.glycan_composition_glyphs.set_transform(transform + self.ax.transData)
        return self

    def make_position_glyph(self, position, i, x, y, size, lw=0):
        glyph = SequencePositionGlyph(position, i, x, y, size=size, lw=lw)
        return glyph

    def render(self):
        ax = self.ax
        if ax is None:
            fig, ax = plt.subplots(1)
            self.ax = ax
        else:
            ax = self.ax

        size = self.size
        x = self.x
        y = self.y
        glyphs = self.sequence_position_glyphs = []
        i = 0
        for position in self.sequence:
            glyph = self.make_position_glyph(position, i, x, y, size=size, lw=self.options.get("lw", 0))
            glyph.render(ax)
            glyphs.append(glyph)
            x += size * self.step_coefficient
            i += 1
        if self.draw_glycan:
            self.glycan_composition_glyphs = GlycanCompositionGlyphs(
                self.sequence.glycan_composition, x + size * self.step_coefficient, y + 0.35, ax
            )
            self.glycan_composition_glyphs.render()
        return ax

    def __getitem__(self, i):
        return self.sequence_position_glyphs[i]

    def __len__(self):
        return len(self.sequence_position_glyphs)

    def next_between(self, index):
        for i, position in enumerate(self.sequence_position_glyphs, 1):
            if i == index:
                break
        mid_point_offset = (self.step_coefficient * self.size) / 1.3
        return position.x + mid_point_offset

    def draw_bar_at(self, index, height=0.25, color="red", **kwargs):
        x = self.next_between(index)
        y = self.y
        rect = mpatch.Rectangle((x, y - height), 0.05, 1 + height, color=color, **kwargs)
        self.ax.add_patch(rect)
        self.annotations.append(rect)

    def draw_n_term_label(self, index, label, height=0.25, length=0.75, size=0.45, color="black", **kwargs):
        kwargs.setdefault("lw", self.options.get("lw", 0))
        x = self.next_between(index)
        y = self.y - height - size
        length *= self.step_coefficient
        label_x = x - length * 0.9
        label_y = y

        if self.next_at_height["n_term"][index] != 0:
            label_y = self.next_at_height["n_term"][index] - 0.4
            self.multi_tier_annotation = True
        else:
            label_y = y

        if self.label_fragments:
            tpath = textpath.TextPath((label_x, label_y), label, size=size, prop=font_options)
            tpatch = mpatch.PathPatch(tpath, color=color, lw=kwargs.get("lw"))
            self.ax.add_patch(tpatch)
            self.annotations.append(tpatch)
            self.next_at_height["n_term"][index] = tpath.vertices[:, 1].min()

    def draw_n_term_annotation(self, index, height=0.25, length=0.75, color="red", **kwargs):
        x = self.next_between(index)
        y = self.y - height
        length *= self.step_coefficient
        rect = mpatch.Rectangle((x - length, y), length, 0.07, color=color, **kwargs)
        self.ax.add_patch(rect)
        self.annotations.append(rect)

    def draw_c_term_label(self, index, label, height=0.25, length=0.75, size=0.45, color="black", **kwargs):
        kwargs.setdefault("lw", self.options.get("lw", 0))
        x = self.next_between(index)
        y = (self.y * 2) + height
        length *= self.step_coefficient
        label_x = x + length / 10.0
        if self.next_at_height["c_term"][index] != 0:
            label_y = self.next_at_height["c_term"][index]
            self.multi_tier_annotation = True
        else:
            label_y = y + 0.2

        if self.label_fragments:
            tpath = textpath.TextPath((label_x, label_y), label, size=size, prop=font_options)
            tpatch = mpatch.PathPatch(tpath, color=color, lw=kwargs.get("lw"))
            self.ax.add_patch(tpatch)
            self.annotations.append(tpatch)
            self.next_at_height["c_term"][index] = tpath.vertices[:, 1].max() + 0.1

    def draw_c_term_annotation(self, index, height=0.0, length=0.75, color="red", **kwargs):
        x = self.next_between(index)
        y = (self.y * 2) + height
        length *= self.step_coefficient
        rect = mpatch.Rectangle((x, y), length, 0.07, color=color, **kwargs)
        self.ax.add_patch(rect)
        self.annotations.append(rect)

    def bbox(self) -> BBox:
        bbox: BBox = None
        for glyph in self.sequence_position_glyphs:
            if bbox is None:
                bbox = glyph.bbox()
            else:
                bbox = bbox.expand(*glyph.bbox())
        if self.glycan_composition_glyphs is not None:
            if bbox is None:
                bbox = self.glycan_composition_glyphs.bbox()
            else:
                bbox = bbox.expand(*self.glycan_composition_glyphs.bbox())
        for glyph in self.annotations:
            if bbox is None:
                bbox = glyph.bbox()
            else:
                bbox = bbox.expand(*glyph.bbox())
        return bbox

    def layout(self):
        ax = self.ax
        # xmax = self.x + self.size * self.step_coefficient * len(self.sequence) + 1
        # if self.draw_glycan:
        #     xmax += self.glycan_composition_glyphs.xend - self.glycan_composition_glyphs.x + 1
        # ax.set_xlim(self.x - 1, xmax)
        # ax.set_ylim(self.y - 1, self.y + 2)
        # if self.multi_tier_annotation:
        #     ax.set_ylim(self.y - 2, self.y + 3)
        bbox = self.bbox()
        ax.set_xlim(bbox.xmin - 0.5, bbox.xmax + 0.5)
        ax.set_ylim(bbox.ymin - 0.5, bbox.ymax + 0.5)
        ax.axis("off")

    def draw(self):
        self.layout()

    def break_at(self, idx):
        if self._fragment_index is None:
            self._build_fragment_index()
        return self._fragment_index[idx]

    def _build_fragment_index(self, types=tuple("bycz")):
        self._fragment_index = [[] for i in range(len(self) + 1)]
        for series in types:
            series = IonSeries(series)
            if series.direction > 0:
                g = self.sequence.get_fragments(series)
                for frags in g:
                    position = self._fragment_index[frags[0].position]
                    position.append(frags)
            else:
                g = self.sequence.get_fragments(series)
                for frags in g:
                    position = self._fragment_index[len(self) - frags[0].position]
                    position.append(frags)

    def build_annotations(
        self, fragments: List[glycopeptidepy.PeptideFragment]
    ) -> Tuple[List[Tuple[glycopeptidepy.PeptideFragment, bool, str]]]:
        index = {}
        for i in range(1, len(self.sequence)):
            for series in self.break_at(i):
                for f in series:
                    index[f.name] = i
        n_annotations = []
        c_annotations = []

        for fragment in fragments:
            if hasattr(fragment, "base_name"):
                key = fragment.base_name()
            else:
                key = fragment.name
            if key in index:
                is_glycosylated = fragment.is_glycosylated
                if key.startswith("b") or key.startswith("c"):
                    n_annotations.append((index[key], is_glycosylated, key))
                elif key.startswith("y") or key.startswith("z"):
                    c_annotations.append((index[key], is_glycosylated, key))
        return n_annotations, c_annotations

    def annotate_from_fragments_stroke(self, fragments: List[glycopeptidepy.PeptideFragment], **kwargs):
        n = len(self.sequence)
        kwargs.setdefault("lw", 2)
        kwargs.setdefault("color", "red")
        kwargs.setdefault("glycosylated_color", "green")
        kwargs.setdefault("v_step", 0.3)
        kwargs.setdefault("stroke_slope", 0)
        kwargs.setdefault("stroke_length", 0.8)
        kwargs.setdefault("glyph_padding", 0.2)

        lw = kwargs["lw"]
        color = kwargs["color"]
        glycosylated_color = kwargs["glycosylated_color"]
        v_step = kwargs["v_step"]
        stroke_slope = kwargs["stroke_slope"]
        stroke_length = kwargs["stroke_length"]
        glyph_padding = kwargs["glyph_padding"]

        annotations = [
            FragmentStroke(
                i,
                self.next_between(i),
                self.y + 0.8 + glyph_padding,
                self.y - 0.1 - glyph_padding,
                n_length=stroke_length,
                c_length=stroke_length,
                v_step=v_step,
                stroke_slope=stroke_slope,
                ax=self.ax,
                main_color=color,
                lw=lw,
            )
            for i in range(1, n)
        ]

        for frag in sorted(fragments, key=lambda x: x.mass):
            if frag.series.includes_peptide and frag.series.direction:
                if frag.series.direction > 0:
                    stroke = annotations[frag.position - 1]
                    if frag.is_glycosylated:
                        if not stroke.has_n_annotations():
                            stroke.add_n_label("", "none")
                        stroke.add_n_label(frag.base_name(), glycosylated_color)
                    else:
                        stroke.add_n_label(frag.base_name(), color)
                else:
                    stroke = annotations[n - (frag.position + 1)]
                    if frag.is_glycosylated:
                        if not stroke.has_c_annotations():
                            stroke.add_c_label("", "none")
                        stroke.add_c_label(frag.base_name(), glycosylated_color)
                    else:
                        stroke.add_c_label(frag.base_name(), color)

        annotations = [annot for annot in annotations if annot.has_annotations()]
        for annot in annotations:
            annot.draw()
        self.annotations.extend(annotations)
        return annotations

    def annotate_from_fragments(self, fragments: List[glycopeptidepy.PeptideFragment], **kwargs):
        n_annotations, c_annotations = self.build_annotations(fragments)

        kwargs.setdefault("height", 0.25)
        kwargs_with_greater_height = kwargs.copy()
        kwargs_with_greater_height["height"] = kwargs["height"] * 2
        kwargs.setdefault("color", "red")

        try:
            kwargs.pop("glycosylated_color")
            kwargs_with_greater_height["color"] = kwargs_with_greater_height["glycosylated_color"]
            kwargs_with_greater_height.pop("glycosylated_color")
        except KeyError:
            color = kwargs.get("color", "red")
            try:
                color = cnames.get(color, color)
                rgb = hex2color(color)
            except Exception:
                rgb = color
            kwargs_with_greater_height["color"] = darken(rgb)

        heights_at = defaultdict(float)

        for n_annot, is_glycosylated, key in n_annotations:
            self.draw_bar_at(n_annot, color=kwargs["color"])
            if is_glycosylated:
                self.draw_n_term_annotation(n_annot, **kwargs_with_greater_height)
                if heights_at[n_annot] < kwargs_with_greater_height["height"]:
                    heights_at[n_annot] = kwargs_with_greater_height["height"]
            else:
                self.draw_n_term_annotation(n_annot, label=key, **kwargs)
                if heights_at[n_annot] < kwargs["height"]:
                    heights_at[n_annot] = kwargs["height"]

        labeled = set()
        for n_annot, is_glycosylated, key in n_annotations:
            label = key.split("+")[0]
            if label not in labeled:
                self.draw_n_term_label(n_annot, label=label, height=heights_at[n_annot])
                labeled.add(label)

        kwargs_with_greater_height["height"] = kwargs.get("height", 0.25)
        kwargs["height"] = 0

        heights_at = defaultdict(float)

        for c_annot, is_glycosylated, key in c_annotations:
            self.draw_bar_at(c_annot, color=kwargs["color"])
            if is_glycosylated:
                self.draw_c_term_annotation(c_annot, **kwargs_with_greater_height)
                if heights_at[c_annot] < kwargs_with_greater_height["height"]:
                    heights_at[c_annot] = kwargs_with_greater_height["height"]
            else:
                self.draw_c_term_annotation(c_annot, label=key, **kwargs)
                if heights_at[c_annot] < kwargs["height"]:
                    heights_at[c_annot] = kwargs["height"]

        labeled = set()
        for c_annot, is_glycosylated, key in c_annotations:
            label = key.split("+")[0]
            if label not in labeled:
                self.draw_c_term_label(c_annot, label=label, height=heights_at[c_annot])
                labeled.add(label)

    @classmethod
    def from_spectrum_match(
        cls, spectrum_match, ax=None, set_layout=True, color="red", glycosylated_color="forestgreen", **kwargs
    ):
        annotation_options = kwargs.get("annotation_options", {})
        annotation_options["color"] = color
        annotation_options["glycosylated_color"] = glycosylated_color
        kwargs["annotation_options"] = annotation_options
        inst = cls(spectrum_match.target, ax=ax, **kwargs)
        fragments = [f for f in spectrum_match.solution_map.fragments()]
        # inst.annotate_from_fragments(fragments, **annotation_options)
        inst.annotate_from_fragments_stroke(fragments, **annotation_options)
        if set_layout:
            inst.layout()
        return inst


def glycopeptide_match_logo(
    glycopeptide_match,
    ax=None,
    color="red",
    glycosylated_color="forestgreen",
    return_artist: bool = True,
    annotation_options: Optional[dict] = None,
    **kwargs,
):
    if annotation_options is None:
        annotation_options = {}
    annotation_options["color"] = color
    annotation_options["glycosylated_color"] = glycosylated_color
    kwargs["annotation_options"] = annotation_options
    inst = SequenceGlyph.from_spectrum_match(
        glycopeptide_match, color=color, glycosylated_color=glycosylated_color, ax=ax, **kwargs
    )
    if return_artist:
        return inst
    return inst.ax


class ProteoformDisplay(object):
    def __init__(self, proteoform, ax=None, width=60):
        self.proteoform = proteoform
        self.glyph_rows = []
        self.peptide_rows = []
        self.width = width
        self.x = 1
        self.y = 1


def draw_proteoform(proteoform, width=60):
    """Draw a sequence logo for a proteoform"""
    fig, ax = plt.subplots(1)
    i = 0
    n = len(proteoform)
    y = 1
    x = 1
    rows = []
    if width > n:
        width = n + 1
    # partion the protein sequence into chunks of width positions
    while i < n:
        rows.append(glycopeptidepy.PeptideSequence.from_iterable(proteoform[i : i + width]))
        i += width
    # draw each row, starting from the last row and going up
    # so that if the current row has a glycan on it, we can
    # compute its height and start drawing the next row further
    # up the plot
    for row in rows[::-1]:
        sg = SequenceGlyph(row, x=x, y=y, ax=ax)
        sg.layout()
        drawn_trees = []
        for position in sg.sequence_position_glyphs:
            # check for glycosylation at each position
            if position.position[1] and position.position[1][0].is_a(modification.ModificationCategory.glycosylation):
                mod = position.position[1][0]
                dtree, ax = draw_tree(
                    mod.rule._original, at=(position.x, position.y + 1), orientation="vertical", ax=sg.ax, center=False
                )
                # re-center tree because draw_tree's layout only sets the starting
                # coordinates of the tree root, and later transforms move it around.
                # this sequence of translates will re-center it above the residue
                dtree.transform.translate(-dtree.x, 0).translate(0.3, 0)
                drawn_trees.append(dtree)
        # get the height of the tallest glycan
        if drawn_trees:
            tree_height = max(dtree.extrema()[3] for dtree in drawn_trees) + 0.5
        else:
            tree_height = 0
        ymax = tree_height + 1
        y += ymax
    ax.set_xlim(0, width + 1)
    ax.set_ylim(-1, y + 1)
    return ax



from typing import Any
import urllib
import logging
import base64

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

from io import BytesIO

from six import string_types as basestring

from matplotlib.axes import Axes
from matplotlib import pyplot as plt
from matplotlib import rcParams as mpl_params

from lxml import etree

import jinja2
from markupsafe import escape

from glypy.structure.glycan_composition import GlycanComposition
from glycopeptidepy import PeptideSequence

from glycresoft.task import TaskBase
from glycresoft.serialize import DatabaseBoundOperation
from glycresoft.symbolic_expression import GlycanSymbolContext
from glycresoft.plotting import colors


mpl_params.update({
    'figure.facecolor': 'white',
    'figure.edgecolor': 'white',
    'font.size': 10,
    'savefig.dpi': 72,
    'figure.subplot.bottom': .125})


status_logger = logging.getLogger("glycresoft.status")


def _sanitize_savefig_kwargs(kwargs: dict):
    keys = ["patchless", "img_width", "width", "height", "svg_width"]
    for key in keys:
        kwargs.pop(key, None)
    return kwargs



def render_plot(figure, **kwargs):
    if isinstance(figure, Axes):
        figure = figure.get_figure()
    if "height" in kwargs:
        figure.set_figheight(kwargs["height"])
    if "width" in kwargs:
        figure.set_figwidth(kwargs['width'])
    if kwargs.get("bbox_inches") != 'tight' or kwargs.get("patchless"):
        figure.patch.set_alpha(0)
        figure.axes[0].patch.set_alpha(0)
    data_buffer = BytesIO()
    figure.savefig(data_buffer, **_sanitize_savefig_kwargs(kwargs))
    plt.close(figure)
    return data_buffer


def xmlattrs(**kwargs):
    return ' '.join("%s=\"%s\"" % kv for kv in kwargs.items()).encode('utf8')


def png_plot(figure, img_width=None, img_height=None, xattrs=None, optimize: bool=False, **kwargs):
    if xattrs is None:
        xattrs = dict()
    kwargs.pop("svg_width", None)
    xml_attributes = dict(xattrs)
    if img_width is not None:
        xml_attributes['width'] = img_width
    if img_height is not None:
        xml_attributes['height'] = img_height
    data_buffer = render_plot(figure, format='png', **
                              _sanitize_savefig_kwargs(kwargs))
    return (b"<img %s src='data:image/png;base64,%s'>" % (
        xmlattrs(**xml_attributes),
        base64.b64encode(data_buffer.getvalue())
        )).decode('utf8')


def svguri_plot(figure, img_width=None, img_height=None, xattrs=None, optimize: bool=False, **kwargs):
    if xattrs is None:
        xattrs = dict()
    xml_attributes = dict(xattrs)
    if img_width is not None:
        xml_attributes['width'] = img_width
    if img_height is not None:
        xml_attributes['height'] = img_height
    svg_string = svg_plot(figure, optimize=optimize, **kwargs)
    return ("<img %s src='data:image/svg+xml;utf-8,%s'>" % (xmlattrs(**xml_attributes).decode('utf8'),
                                                            quote(svg_string)))


def _strip_style(root):
    style_node = root.find(".//{http://www.w3.org/2000/svg}style")
    style_node.text = ""
    return root


def svg_plot(figure, svg_width=None, xml_transform=None, optimize: bool=False, **kwargs):
    data_buffer = render_plot(figure, format='svg', **
                              _sanitize_savefig_kwargs(kwargs))
    if optimize:
        root = etree.fromstring(svg_optimizer(data_buffer.getvalue()))
    else:
        root = etree.fromstring(data_buffer.getvalue())
    root = _strip_style(root)
    if svg_width is not None:
        root.attrib["width"] = svg_width
    if xml_transform is not None:
        root = xml_transform(root)
    return etree.tostring(root)


def rgbpack(color):
    return "rgba(%d,%d,%d,0.5)" % tuple(i * 255 for i in color)


def formula(composition):
    return ''.join("<b>%s</b><sub>%d</sub>" % (k, v) for k, v in sorted(composition.items()))


def modification_specs(modification_rule):
    if isinstance(modification_rule, str):
        yield modification_rule
    else:
        yield from modification_rule.as_spec_strings()


def glycan_composition_string(composition):
    try:
        composition = GlycanComposition.parse(
            GlycanSymbolContext(
                GlycanComposition.parse(
                    composition)).serialize())
    except ValueError:
        return "<code>%s</code>" % composition

    parts = []
    template = ("<span class='monosaccharide-composition-name'"
                "style='background-color:%s'>"
                "%s&nbsp;%d</span>")
    for k, v in sorted(composition.items(), key=lambda x: x[0].mass()):
        name = str(k)
        color = colors.get_color(str(name))
        parts.append(template % (rgbpack(color), name, v))
    reduced = composition.reducing_end
    if reduced:
        reducing_end_template = (
            "<span class='monosaccharide-composition-name'"
            "style='background-color:%s;padding:2px;border-radius:2px;'>"
            "%s</span>")
        name = formula(reduced.composition)
        color = colors.get_color(str(name))
        parts.append(reducing_end_template % (rgbpack(color), name))

    return ' '.join(parts)


def glycopeptide_string(sequence, long=False, include_glycan=True):
    sequence = PeptideSequence(str(sequence))
    parts = []
    template = "(<span class='modification-chip'"\
        " style='background-color:%s;padding-left:1px;padding-right:2px;border-radius:2px;'"\
        " title='%s' data-modification='%s'>%s</span>)"

    n_term_template = template.replace("(", "").replace(")", "") + '-'
    c_term_template = "-" + (template.replace("(", "").replace(")", ""))

    def render(mod, template=template):
        color = colors.get_color(str(mod))
        letter = escape(mod.name if long else mod.name[0])
        name = escape(mod.name)
        parts.append(template % (rgbpack(color), name, name, letter))

    if sequence.n_term.modification is not None:
        render(sequence.n_term.modification, n_term_template)
    for res, mods in sequence:
        parts.append(res.symbol)
        if mods:
            for mod in mods:
                render(mod)
    if sequence.c_term.modification is not None:
        render(sequence.c_term.modification, c_term_template)
    parts.append((
        ' ' + glycan_composition_string(str(sequence.glycan)) if sequence.glycan is not None else "")
        if include_glycan else "")
    return ''.join(parts)


def highlight_sequence_site(amino_acid_sequence, site_list, site_type_list):
    if isinstance(site_type_list, basestring):
        site_type_list = [site_type_list for i in site_list]
    sequence = list(amino_acid_sequence)
    for site, site_type in zip(site_list, site_type_list):
        sequence[site] = "<span class='{}'>{}</span>".format(site_type, sequence[site])
    return sequence


def n_per_row(sequence, n=60):
    row_buffer = []
    i = 0
    while i < len(sequence):
        row_buffer.append(
            ''.join(sequence[i:(i + n)])
        )
        i += n
    return '<br>'.join(row_buffer)


def is_type(x):
    return isinstance(x, type)


class ReportCreatorBase(TaskBase):
    database_connection: DatabaseBoundOperation
    env: jinja2.Environment

    def __init__(self, database_connection, analysis_id, stream=None):
        self.database_connection = DatabaseBoundOperation(database_connection)
        self.analysis_id = analysis_id
        self.stream = stream
        self.env = jinja2.Environment()

    @property
    def session(self):
        return self.database_connection.session

    def status_update(self, message):
        status_logger.info(message)

    def prepare_environment(self):
        self.env.filters['svguri_plot'] = svguri_plot
        self.env.filters['png_plot'] = png_plot
        self.env.filters['svg_plot'] = svg_plot
        self.env.filters["n_per_row"] = n_per_row
        self.env.filters['highlight_sequence_site'] = highlight_sequence_site
        self.env.filters['svg_plot'] = svg_plot
        self.env.filters['glycopeptide_string'] = glycopeptide_string
        self.env.filters['glycan_composition_string'] = glycan_composition_string
        self.env.filters["formula"] = formula
        self.env.filters["modification_specs"] = modification_specs
        self.env.tests['is_type'] = is_type

    def set_template_loader(self, path):
        self.env.loader = jinja2.FileSystemLoader(path)

    def make_template_stream(self):
        raise NotImplementedError()

    def run(self):
        self.prepare_environment()
        template_stream = self.make_template_stream()
        template_stream.dump(self.stream, encoding='utf-8')

try:
    from scour import scour

    class _SVGOptimizer:
        def __init__(self, **parameters):
            self.parameters = parameters
            self.scour_args = scour.generateDefaultOptions()
            for k, v in parameters.items():
                setattr(self.scour_args, k, v)

        def __call__(self, xml_string: bytes) -> bytes:
            return scour.scourString(xml_string, self.scour_args).encode('utf8')

except ImportError:
    class _SVGOptimizer:
        def __init__(self, **parameters) -> None:
            self.parameters = parameters

        def __call__(self, xml_string: bytes) -> bytes:
            if isinstance(xml_string, str):
                return xml_string.encode('utf8')
            return xml_string


svg_optimizer = _SVGOptimizer(
    strip_comments=True,
    strip_ids=True,
    shorten_ids=True,
    remove_metadata=True,
    indent_depth=0
)



import numpy as np
from matplotlib import pyplot as plt
from glycresoft.composition_distribution_model import laplacian_matrix


def _spectral_layout(network):
    eigvals, eigvects = np.linalg.eig(laplacian_matrix(network))
    index = np.argsort(eigvects)[2:3]
    coords = (eigvects[:, index[0][0]], eigvects[:, index[0][1]])
    return coords


def draw_network(network, ax=None, layout='spectral'):
    if ax is None:
        fig, ax = plt.subplots(1)
    font_args = {"ha": "center", 'va': 'center', "fontdict": {"family": "monospace"}}

    # Spectral Layout Algorithm
    if layout == 'spectral':
        coords = _spectral_layout(network)
    elif layout == 'spring':
        coords = _spring_layout(network)

    seen_edges = set()
    for node in network.nodes:
        xc, yc = coords[0][node.index], coords[1][node.index]
        ax.scatter(xc, yc, s=250, alpha=0.5)
        label = str(node)
        label = '\n'.join([t.replace(":", "") for t in label[1:-1].split("; ")])
        ax.text(xc, yc, label, size=10, **font_args)
        for e in node.edges:
            if e in seen_edges:
                continue
            else:
                seen_edges.add(e)
            neighbor = e[node]
            neigh_xc, neigh_yc = coords[0][neighbor.index], coords[1][neighbor.index]
            ax.plot((xc, neigh_xc), (yc, neigh_yc), color='grey', alpha=0.5)
            center_xc = (xc + neigh_xc) / 2.
            center_yc = (yc + neigh_yc) / 2.
            ax.text(center_xc, center_yc, "%0.2f" % (1. / e.order), size=7, **font_args)
    ax.axis("off")
    return ax


class VertexWrapper(object):
    def __init__(self, node, position, dispersion):
        self.node = node
        self.position = position
        self.dispersion = dispersion

    def __eq__(self, other):
        return self.node == other.node

    def __ne__(self, other):
        return self.node != other.node


def _spring_layout(network, width=None, height=None, maxiter=3):
    if width is None:
        width = 50.0
    if height is None:
        height = 50.0
    area = width * height
    area = float(area)
    vertices = []
    X, Y = np.random.random(len(network)), np.random.random(len(network))
    temperature = 100.0
    iters = 0
    while iters < maxiter:
        for i, node in enumerate(network.nodes):
            vert = VertexWrapper(node, np.array((X[i], Y[i])), 0)
            vertices.append(vert)
        k = np.sqrt(area / i)

        def attract(x):
            return (x * x) / k

        def repel(x):
            return (k * k) / x

        for v in vertices:
            v.dispersion = 0
            for u in vertices:
                if v == u:
                    continue
                delta = v.position - u.position
                magdelta = len(delta)
                v.dispersion = v.dispersion + (delta / magdelta) * repel(magdelta)

        for edge in network.edges:
            v = vertices[edge.node1.index]
            u = vertices[edge.node2.index]
            delta = v.position - u.position
            magdelta = len(delta)
            attract_magdelta = attract(magdelta)
            delta_magdelta_ratio = (delta / magdelta)
            dispersion_shift = delta_magdelta_ratio * attract_magdelta

            v.dispersion = v.dispersion - dispersion_shift
            u.dispersion = u.dispersion + dispersion_shift

        for v in vertices:
            v.position = v.position + (v.dispersion / len(v.dispersion)) * temperature
            v.position /= 1e4
        temperature = max(temperature / 2., 1)
        iters += 1
    X = (np.array([v.position[0] for v in vertices]) + 5) * 10
    Y = (np.array([v.position[1] for v in vertices]) + 5) * 10
    return X, Y



import click

from glycresoft.cli.base import cli
from glycresoft.config.config_file import (
    add_user_modification_rule as add_user_peptide_modification_rule,
    add_user_substituent_rule, delete_config_dir)

from glypy.composition import formula, Composition

from glycopeptidepy.structure.modification import (
    extract_targets_from_string, ModificationRule,
    Modification)


@cli.group(short_help='Set persistent configuration options')
def config():
    pass


def parse_peptide_modification_target_spec(context, spec_string_list):
    out = []
    for spec_string in spec_string_list:
        try:
            spec = extract_targets_from_string(spec_string)
            out.append(spec)
        except Exception as e:
            raise ValueError(str(e))
    return out


@config.command("add-peptide-modification")
@click.option("-n", '--name', required=True, help='Modification name')
@click.option("-c", "--composition", required=True, help='The chemical formula for this modification')
@click.option("-t", "--target", required=True, multiple=True,
              help="Target specification string of the form residue[@n-term|c-term]",
              callback=parse_peptide_modification_target_spec)
def peptide_modification(name, composition, target, categories=None):
    composition = Composition(str(composition))
    rule = ModificationRule(target, name, None, composition.mass, composition)
    add_user_peptide_modification_rule(rule)
    click.echo("Added %r to modification registry" % (rule,))


@config.command("get-peptide-modification")
@click.argument("name")
def display_peptide_modification(name):
    mod = Modification(name)
    click.echo("name: %s" % mod.name)
    click.echo("mass: %f" % mod.mass)
    click.echo("formula: %s" % formula(mod.composition))
    for target in mod.rule.targets:
        click.echo("target: %s" % target.serialize())


@config.command("add-substituent")
@click.option("-n", "--name", required=True, help='Substituent name')
@click.option("-c", "--composition", required=True, help='The chemical formula for this substituent')
@click.option("-s", "--is-nh-derivatizable", is_flag=True,
              help="Can this substituent be derivatized as at an N-H bond?")
@click.option("-d", "--can-nh-derivatize", is_flag=True,
              help="Will this substituent derivatize other substituents at sites like an N-H bond?")
@click.option("-a", "--attachment-loss", default="H",
              help="The composition lost by the parent molecule when this substituent is added. Defaults to \"H\"")
def substituent(name, composition, is_nh_derivatizable, can_nh_derivatize, attachment_loss):
    click.echo("name: %s" % (name,))
    click.echo("composition: %s" % (composition,))
    click.echo("is_nh_derivatizable: %s" % (is_nh_derivatizable,))
    click.echo("can_nh_derivatize: %s" % (can_nh_derivatize,))
    click.echo("attachment_loss: %s" % (attachment_loss,))


@config.command("clear-configuration")
def clear_config():
    click.secho("Deleting Configuration Directory", fg='red')
    delete_config_dir()



import warnings

from typing import Any, DefaultDict, Dict, List, Optional, Tuple, Deque

from sqlalchemy import (
    Column, Numeric, Integer, String, ForeignKey,
    Boolean, Table, select)
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import (
    relationship, backref, object_session, Load)
from sqlalchemy.exc import OperationalError

from glycresoft.tandem.spectrum_match import (
    SpectrumMatch as MemorySpectrumMatch,
    SpectrumSolutionSet as MemorySpectrumSolutionSet,
    ScoreSet, FDRSet, MultiScoreSpectrumMatch,
    MultiScoreSpectrumSolutionSet, LocalizationScore as MemoryLocalizationScore)

from glycresoft.structure import ScanInformation

from .analysis import BoundToAnalysis
from .hypothesis import Glycopeptide, GlycanComposition
from .chromatogram import CompoundMassShift
from .spectrum import (
    Base, MSScan)
from .utils import chunkiter, needs_migration

def convert_scan_to_record(scan):
    return ScanInformation(
        scan.scan_id, scan.index, scan.scan_time,
        scan.ms_level, scan.precursor_information.convert())


class SpectrumMatchBase(BoundToAnalysis):

    score = Column(Numeric(12, 6, asdecimal=False), index=True)

    @declared_attr
    def scan_id(self):
        return Column(Integer, ForeignKey(MSScan.id), index=True)

    @declared_attr
    def scan(self):
        return relationship(MSScan)

    @declared_attr
    def mass_shift_id(self):
        return needs_migration(Column(Integer, ForeignKey(CompoundMassShift.id), index=True))

    @declared_attr
    def mass_shift(self):
        return relationship(CompoundMassShift, uselist=False)

    def _check_mass_shift(self, session):
        flag = session.info.get("has_spectrum_match_mass_shift")
        if flag:
            return self.mass_shift_id
        elif flag is None:
            try:
                session.begin_nested()
                shift_id = self.mass_shift_id
                session.rollback()
                session.info["has_spectrum_match_mass_shift"] = True
                return shift_id
            except OperationalError as err:
                session.rollback()
                warnings.warn(
                    ("Encountered an error while checking if "
                     "SpectrumMatch-tables support mass shifts: %r") % err)
                session.info["has_spectrum_match_mass_shift"] = False
                return None
        else:
            return None

    def _resolve_mass_shift(self, session, mass_shift_cache=None):
        mass_shift_id = self._check_mass_shift(session)
        if mass_shift_id is not None:
            if mass_shift_cache is not None:
                try:
                    mass_shift = mass_shift_cache[mass_shift_id]
                except KeyError:
                    mass_shift = self.mass_shift
                    mass_shift_cache[mass_shift_id] = mass_shift.convert()
                    mass_shift = mass_shift_cache[mass_shift_id]
                return mass_shift
            else:
                return self.mass_shift.convert()
        else:
            return None

    def is_multiscore(self):
        """Check whether this match has been produced by summarizing a multi-score
        match, rather than a single score match.

        Returns
        -------
        bool
        """
        return False


class SpectrumClusterBase(object):
    def __getitem__(self, i):
        return self.spectrum_solutions[i]

    def __iter__(self):
        return iter(self.spectrum_solutions)

    def __len__(self):
        return len(self.spectrum_solutions)

    def __repr__(self):
        template = "{self.__class__.__name__}(<{size}>)"
        return template.format(self=self, size=len(self))

    @classmethod
    def compute_key(cls, members) -> frozenset:
        return frozenset([m.key for m in members])

    def convert(self, mass_shift_cache=None, scan_cache=None, structure_cache=None, **kwargs):
        if scan_cache is None:
            scan_cache = dict()
        if mass_shift_cache is None:
            mass_shift_cache = dict()
        matches = [x.convert(mass_shift_cache=mass_shift_cache, scan_cache=scan_cache,
                             structure_cache=structure_cache, **kwargs) for x in self.spectrum_solutions]
        return matches

    def is_multiscore(self) -> bool:
        """Check whether this match collection has been produced by summarizing a multi-score
        match, rather than a single score match.

        Returns
        -------
        bool
        """
        return self[0].is_multiscore()


class SolutionSetBase(object):

    @declared_attr
    def scan_id(self):
        return Column(Integer, ForeignKey(MSScan.id), index=True)

    @declared_attr
    def scan(self):
        return relationship(MSScan)

    @property
    def scan_time(self):
        return self.scan.scan_time

    def best_solution(self) -> SpectrumMatchBase:
        best_match = self.spectrum_matches.filter_by(is_best_match=True).first()
        if best_match is not None:
            # If the best match is already marked, return it
            return best_match
        if not self.is_multiscore():
            return sorted(self.spectrum_matches, key=lambda x: x.score, reverse=True)[0]
        else:
            return sorted(self.spectrum_matches, key=lambda x: (x.score_set.convert()[0]), reverse=True)[0]

    def is_ambiguous(self) -> bool:
        seen = set()

        for sm in self.spectrum_matches.filter_by(is_best_match=True).all():
            seen.add(str(sm.target))
        return len(seen) > 1

    @property
    def key(self) -> frozenset:
        scan_id = self.scan.id
        return frozenset([(scan_id, match.target.id) for match in self])

    @property
    def score(self):
        return self.best_solution().score

    def __getitem__(self, i):
        return self.spectrum_matches[i]

    def __iter__(self):
        return iter(self.spectrum_matches)

    _target_map = None

    def _make_target_map(self):
        self._target_map = {
            sol.target: sol for sol in self
        }

    def solution_for(self, target):
        if self._target_map is None:
            self._make_target_map()
        return self._target_map[target]

    def is_multiscore(self):
        """Check whether this match has been produced by summarizing a multi-score
        match, rather than a single score match.

        Returns
        -------
        bool
        """
        return self[0].is_multiscore()


class GlycopeptideSpectrumCluster(Base, SpectrumClusterBase, BoundToAnalysis):
    __tablename__ = "GlycopeptideSpectrumCluster"

    id = Column(Integer, primary_key=True)

    @classmethod
    def serialize(cls, obj, session, scan_look_up_cache, mass_shift_cache, analysis_id, *args, **kwargs):
        inst = cls()
        session.add(inst)
        session.flush()
        cluster_id = inst.id
        for solution_set in obj.tandem_solutions:
            GlycopeptideSpectrumSolutionSet.serialize(
                solution_set, session, scan_look_up_cache, mass_shift_cache, analysis_id,
                cluster_id, *args, **kwargs)
        return inst


class GlycopeptideSpectrumSolutionSet(Base, SolutionSetBase, BoundToAnalysis):
    __tablename__ = "GlycopeptideSpectrumSolutionSet"

    id = Column(Integer, primary_key=True)
    cluster_id = Column(
        Integer,
        ForeignKey(GlycopeptideSpectrumCluster.id, ondelete="CASCADE"),
        index=True)

    cluster = relationship(GlycopeptideSpectrumCluster, backref=backref("spectrum_solutions", lazy='subquery'))

    is_decoy = Column(Boolean, index=True)

    @classmethod
    def serialize(cls, obj: MemorySpectrumSolutionSet, session, scan_look_up_cache, mass_shift_cache, analysis_id,
                  cluster_id, is_decoy=False, *args, **kwargs):
        obj.rank()
        gpsm = obj.best_solution()
        if gpsm is None:
            gpsm = obj[0]
        if not gpsm.best_match:
            obj.mark_top_solutions()
        inst = cls(
            scan_id=scan_look_up_cache[obj.scan.id],
            is_decoy=is_decoy,
            analysis_id=analysis_id,
            cluster_id=cluster_id)
        session.add(inst)
        session.flush()
        GlycopeptideSpectrumMatch.serialize_bulk(
            obj, session, scan_look_up_cache, mass_shift_cache, analysis_id, inst.id)
        return inst

    def convert(self, mass_shift_cache=None, scan_cache=None, structure_cache=None, peptide_relation_cache=None):
        if scan_cache is None:
            scan_cache = {}
        if mass_shift_cache is None:
            mass_shift_cache = {}
        session = object_session(self)
        flag = session.info.get("has_spectrum_match_mass_shift")
        if flag:
            spectrum_match_q = self.spectrum_matches.options(
                Load(GlycopeptideSpectrumMatch).undefer("mass_shift_id")).all()
        else:
            spectrum_match_q = self.spectrum_matches.all()
        matches = [x.convert(mass_shift_cache=mass_shift_cache, scan_cache=scan_cache,
                             structure_cache=structure_cache, peptide_relation_cache=peptide_relation_cache)
                             for x in spectrum_match_q]
        solution_set_tp = MemorySpectrumSolutionSet
        if matches and matches[0].is_multiscore():
            solution_set_tp = MultiScoreSpectrumSolutionSet
        inst = solution_set_tp(
            convert_scan_to_record(self.scan),
            matches
        )
        inst._is_sorted = True
        inst.id = self.id
        return inst

    def __repr__(self):
        return "DB" + repr(self.convert())

    @classmethod
    def bulk_load(cls, session,
                  ids,
                  chunk_size: int=512,
                  mass_shift_cache=None,
                  scan_cache=None,
                  structure_cache=None,
                  peptide_relation_cache=None,
                  flatten=True,
                  min_q_value=0.2):
        return cls._inner_bulk_load(
            cls.id.in_,
            session,
            ids,
            chunk_size,
            mass_shift_cache,
            scan_cache,
            structure_cache,
            peptide_relation_cache,
            flatten,
            min_q_value=min_q_value
        )

    @classmethod
    def bulk_load_from_clusters(cls, session,
                                ids,
                                chunk_size: int = 512,
                                mass_shift_cache=None,
                                scan_cache=None,
                                structure_cache=None,
                                peptide_relation_cache=None,
                                min_q_value=0.2
                            ):
        by_id = cls._inner_bulk_load(
            cls.cluster_id.in_,
            session,
            ids,
            chunk_size,
            mass_shift_cache,
            scan_cache,
            structure_cache,
            peptide_relation_cache,
            flatten=False,
            min_q_value=min_q_value
        )
        groups = DefaultDict(list)
        for chunk in chunkiter(ids, chunk_size):
            res = session.execute(select([cls.id, cls.cluster_id]).where(cls.cluster_id.in_(chunk)))
            for (sset_id, cluster_id) in res.fetchall():
                groups[cluster_id].append(by_id[sset_id])
        return groups

    @classmethod
    def _inner_bulk_load(
        cls,
        selector,
        session,
        ids,
        chunk_size: int = 512,
        mass_shift_cache=None,
        scan_cache=None,
        structure_cache=None,
        peptide_relation_cache=None,
        flatten=True,
        min_q_value=0.2
    ):
        if selector is None:
            selector = cls.id.in_
        if scan_cache is None:
            scan_cache = {}
        if mass_shift_cache is None:
            mass_shift_cache = {
                m.id: m.convert() for m in session.query(CompoundMassShift).all()
            }
        out = {}
        for chunk in chunkiter(ids, chunk_size):
            res = session.execute(cls.__table__.select().where(
                selector(chunk)))
            for self in res.fetchall():
                gpsm_ids = [
                    i[0] for i in session.query(GlycopeptideSpectrumMatch.id).filter(
                        GlycopeptideSpectrumMatch.solution_set_id == self.id,
                        GlycopeptideSpectrumMatch.q_value <= min_q_value
                    ).all()
                ]
                if not gpsm_ids:
                    gpsm_ids = [
                        i[0] for i in session.query(GlycopeptideSpectrumMatch.id).filter(
                            GlycopeptideSpectrumMatch.solution_set_id == self.id,
                        ).order_by(
                            GlycopeptideSpectrumMatch.q_value.asc(),
                            GlycopeptideSpectrumMatch.score.desc()
                        ).limit(5)
                    ]
                matches = GlycopeptideSpectrumMatch.bulk_load(
                    session,
                    gpsm_ids,
                    chunk_size,
                    mass_shift_cache=mass_shift_cache,
                    scan_cache=scan_cache,
                    structure_cache=structure_cache,
                    peptide_relation_cache=peptide_relation_cache
                )
                matches = [matches[k] for k in gpsm_ids]
                solution_set_tp = MemorySpectrumSolutionSet
                if matches and matches[0].is_multiscore():
                    solution_set_tp = MultiScoreSpectrumSolutionSet

                if self.scan_id not in scan_cache:
                    MSScan.bulk_load(session, [self.scan_id], scan_cache=scan_cache)
                inst = solution_set_tp(scan_cache[self.scan_id], matches)
                inst._is_sorted = True
                inst.id = self.id
                out[self.id] = inst
        if flatten:
            return [out[i] for i in ids]
        return out


class GlycopeptideSpectrumMatch(Base, SpectrumMatchBase):
    __tablename__ = "GlycopeptideSpectrumMatch"

    id = Column(Integer, primary_key=True)
    solution_set_id = Column(
        Integer, ForeignKey(
            GlycopeptideSpectrumSolutionSet.id, ondelete='CASCADE'),
        index=True)
    solution_set = relationship(
        GlycopeptideSpectrumSolutionSet, backref=backref("spectrum_matches", lazy='dynamic'))

    q_value = Column(Numeric(8, 7, asdecimal=False), index=True)
    is_decoy = Column(Boolean)
    is_best_match = Column(Boolean, index=True)
    rank = needs_migration(Column(Integer, default=0), 0)

    structure_id = Column(
        Integer, ForeignKey(Glycopeptide.id, ondelete='CASCADE'),
        index=True)

    structure = relationship(Glycopeptide)

    @property
    def target(self):
        return self.structure

    @property
    def q_value_set(self):
        return self.score_set

    def is_multiscore(self):
        """Check whether this match has been produced by summarizing a multi-score
        match, rather than a single score match.

        Returns
        -------
        bool
        """
        return self.score_set is not None

    @classmethod
    def serialize(cls, obj, session, scan_look_up_cache, mass_shift_cache, analysis_id,
                  solution_set_id, is_decoy=False, save_score_set=True, *args, **kwargs):
        inst = cls(
            scan_id=scan_look_up_cache[obj.scan.id],
            is_decoy=is_decoy,
            analysis_id=analysis_id,
            score=obj.score,
            rank=obj.rank,
            q_value=obj.q_value,
            solution_set_id=solution_set_id,
            is_best_match=obj.best_match,
            structure_id=obj.target.id,
            mass_shift_id=mass_shift_cache[obj.mass_shift].id)
        session.add(inst)
        session.flush()
        if hasattr(obj, 'score_set') and save_score_set:
            assert inst.id is not None
            GlycopeptideSpectrumMatchScoreSet.serialize_from_spectrum_match(obj, session, inst.id)
            if obj.localizations:
                locs = (LocalizationScore.get_fields_from_object(obj, inst.id))
                if locs:
                    session.bulk_insert_mappings(LocalizationScore, locs)

        return inst

    @classmethod
    def serialize_bulk(cls, objs, session, scan_look_up_cache, mass_shift_cache, analysis_id,
                       solution_set_id, is_decoy=False, save_score_set=True, *args, **kwargs):
        acc = []
        revmap: Dict[Any, MemorySpectrumMatch] = {}
        for obj in objs:
            scan_id = scan_look_up_cache[obj.scan.id]
            target_id = obj.target.id
            shift_id = mass_shift_cache[obj.mass_shift].id
            inst = cls(
                scan_id=scan_look_up_cache[obj.scan.id],
                is_decoy=is_decoy,
                analysis_id=analysis_id,
                score=obj.score,
                rank=obj.rank,
                q_value=obj.q_value,
                solution_set_id=solution_set_id,
                is_best_match=obj.best_match,
                structure_id=target_id,
                mass_shift_id=shift_id)
            acc.append(inst)
            revmap[scan_id, target_id, shift_id] = obj

        session.bulk_save_objects(acc)

        if save_score_set:
            session.flush()
            fwd = session.query(cls.id, cls.scan_id, cls.structure_id, cls.mass_shift_id).filter(
                cls.solution_set_id == solution_set_id).all()
            acc = []
            loc_acc = []
            for inst in fwd:
                obj = revmap[inst.scan_id, inst.structure_id, inst.mass_shift_id]
                if hasattr(obj, 'score_set'):
                    fields = GlycopeptideSpectrumMatchScoreSet.get_fields_from_object(obj, inst.id)
                    acc.append(fields)
                if obj.localizations:
                    loc_acc.extend(LocalizationScore.get_fields_from_object(obj, inst.id))
            if acc:
                session.bulk_insert_mappings(GlycopeptideSpectrumMatchScoreSet, acc)
            if loc_acc:
                session.bulk_insert_mappings(
                    LocalizationScore,
                    loc_acc
                )

        return revmap

    @classmethod
    def bulk_load(cls, session, ids, chunk_size: int = 512,
                  mass_shift_cache: Optional[Dict] = None,
                  scan_cache: Optional[Dict] = None,
                  structure_cache: Optional[Dict] = None,
                  peptide_relation_cache: Optional[Dict] = None) -> Dict[int, MemorySpectrumMatch]:
        if peptide_relation_cache is None:
            peptide_relation_cache = {}
        if structure_cache is None:
            structure_cache = {}
        if mass_shift_cache is None:
            mass_shift_cache = {m.id: m.convert() for m in session.query(CompoundMassShift).all()}
        if scan_cache is None:
            scan_cache = {}

        out = {}
        id_series = []
        scan_series = []
        mass_shift_series = []
        structure_series = []
        is_best_match_series = Deque()
        rank_series = Deque()
        solution_set_id = Deque()
        score_map = {}

        for chunk in chunkiter(ids, chunk_size):
            res = session.execute(cls.__table__.select().where(
                cls.id.in_(chunk)))
            for self in res.fetchall():
                id_series.append(self.id)
                scan_series.append(self.scan_id)
                mass_shift_series.append(mass_shift_cache[self.mass_shift_id])
                structure_series.append(self.structure_id)
                is_best_match_series.append(self.is_best_match)
                rank_series.append(self.rank)
                solution_set_id.append(self.solution_set_id)
                score_map[self.id] = (self.score, self.q_value)

        solution_set_id_to_cluster_id = {}
        uniq_set_ids = list(set(solution_set_id))
        for chunk in chunkiter(uniq_set_ids, chunk_size):
            res = session.execute(
                select([GlycopeptideSpectrumSolutionSet.id,
                        GlycopeptideSpectrumSolutionSet.cluster_id]).where(
                GlycopeptideSpectrumSolutionSet.id.in_(chunk)))
            for (ssid, clid) in res.fetchall():
                solution_set_id_to_cluster_id[ssid] = clid

        scan_objects = MSScan.bulk_load(
            session, scan_series,
            chunk_size=chunk_size, scan_cache=scan_cache)

        structure_objects = Glycopeptide.bulk_load(
            session, structure_series,
            chunk_size=chunk_size,
            peptide_relation_cache=peptide_relation_cache,
            structure_cache=structure_cache)

        score_sets_map = GlycopeptideSpectrumMatchScoreSet.bulk_load(
            session, ids, chunk_size=chunk_size)

        localization_map = LocalizationScore.bulk_load(session, ids, chunk_size=chunk_size)

        for id, scan, structure, mass_shift, best_match, rank, sset_id in zip(
                id_series,
                scan_objects,
                structure_objects,
                mass_shift_series,
                is_best_match_series,
                rank_series,
                solution_set_id):
            localizations = localization_map[id]
            if id in score_sets_map:
                score_set, fdr_set = score_sets_map[id]
                out[id] = MultiScoreSpectrumMatch(
                    scan, structure,
                    id=id,
                    score_set=score_set,
                    best_match=best_match,
                    q_value_set=fdr_set,
                    mass_shift=mass_shift,
                    localizations=localizations,
                    rank=rank,
                    cluster_id=solution_set_id_to_cluster_id[sset_id]
                )
            else:
                score, q_value = score_map[id]
                out[id] = MemorySpectrumMatch(
                    scan, structure, score, best_match,
                    q_value=q_value,
                    id=id,
                    mass_shift=mass_shift,
                    localizations=localizations,
                    rank=rank,
                    cluster_id=solution_set_id_to_cluster_id[sset_id]
                )
        return out

    def convert(self, mass_shift_cache=None, scan_cache=None, structure_cache=None, peptide_relation_cache=None):
        session = object_session(self)
        key = self.scan_id
        if scan_cache is None:
            scan = session.query(MSScan).get(key).convert(False, False)
        else:
            try:
                scan = scan_cache[key]
            except KeyError:
                scan = scan_cache[key] = session.query(MSScan).get(key).convert(False, False)

        key = self.structure_id
        if structure_cache is None:
            target = self.structure.convert(peptide_relation_cache=peptide_relation_cache)
        else:
            try:
                target = structure_cache[key]
            except KeyError:
                target = structure_cache[key] = self.structure.convert(peptide_relation_cache=peptide_relation_cache)

        orm_locs = self.localizations
        if orm_locs:
            locs = [loc.convert() for loc in orm_locs]
        else:
            locs = []
        mass_shift = self._resolve_mass_shift(session, mass_shift_cache)
        orm_score_set = self.score_set
        if orm_score_set is None:
            inst = MemorySpectrumMatch(scan, target, self.score, self.is_best_match,
                                       mass_shift=mass_shift, localizations=locs)
        else:
            score_set, fdr_set = orm_score_set.convert()
            inst = MultiScoreSpectrumMatch(
                scan, target, score_set, self.is_best_match,
                mass_shift=mass_shift, q_value_set=fdr_set, match_type=0,
                localizations=locs)

        inst.q_value = self.q_value
        inst.id = self.id
        return inst

    def __repr__(self):
        return "DB" + repr(self.convert())

    @property
    def cluster_id(self):
        return self.solution_set.cluster_id


class GlycanCompositionSpectrumCluster(Base, SpectrumClusterBase, BoundToAnalysis):
    __tablename__ = "GlycanCompositionSpectrumCluster"

    id = Column(Integer, primary_key=True)

    @classmethod
    def serialize(cls, obj, session, scan_look_up_cache, mass_shift_cache, analysis_id, *args, **kwargs):
        inst = cls()
        session.add(inst)
        session.flush()
        cluster_id = inst.id
        for solution_set in obj.tandem_solutions:
            GlycanCompositionSpectrumSolutionSet.serialize(
                solution_set, session, scan_look_up_cache, mass_shift_cache, analysis_id,
                cluster_id, *args, **kwargs)
        return inst

    source = relationship(
        "GlycanCompositionChromatogram",
        secondary=lambda: GlycanCompositionChromatogramToGlycanCompositionSpectrumCluster,
        backref=backref("spectrum_cluster", uselist=False))


class GlycanCompositionSpectrumSolutionSet(Base, SolutionSetBase, BoundToAnalysis):
    __tablename__ = "GlycanCompositionSpectrumSolutionSet"

    id = Column(Integer, primary_key=True)
    cluster_id = Column(
        Integer,
        ForeignKey(GlycanCompositionSpectrumCluster.id, ondelete="CASCADE"),
        index=True)

    cluster = relationship(GlycanCompositionSpectrumCluster, backref=backref(
        "spectrum_solutions", lazy='subquery'))

    @classmethod
    def serialize(cls, obj, session, scan_look_up_cache, mass_shift_cache, analysis_id, cluster_id, *args, **kwargs):
        inst = cls(
            scan_id=scan_look_up_cache[obj.scan.id],
            analysis_id=analysis_id,
            cluster_id=cluster_id)
        session.add(inst)
        session.flush()
        # if we have a real SpectrumSolutionSet, then it will be iterable
        try:
            list(obj)
        except TypeError:
            # otherwise we have a single SpectrumMatch
            obj = [obj]
        for solution in obj:
            GlycanCompositionSpectrumMatch.serialize(
                solution, session, scan_look_up_cache, mass_shift_cache,
                analysis_id, inst.id, *args, **kwargs)
        return inst

    def convert(self, mass_shift_cache=None, scan_cache=None, structure_cache=None):
        if scan_cache is None:
            scan_cache = dict()
        matches = [x.convert(mass_shift_cache=mass_shift_cache, scan_cache=scan_cache,
                             structure_cache=structure_cache) for x in self.spectrum_matches]
        matches.sort(key=lambda x: x.score, reverse=True)
        inst = MemorySpectrumSolutionSet(
            convert_scan_to_record(self.scan),
            matches
        )
        inst.q_value = min(x.q_value for x in inst)
        inst.id = self.id
        return inst

    def __repr__(self):
        return "DB" + repr(self.convert())


class GlycanCompositionSpectrumMatch(Base, SpectrumMatchBase):
    __tablename__ = "GlycanCompositionSpectrumMatch"

    id = Column(Integer, primary_key=True)
    solution_set_id = Column(
        Integer, ForeignKey(
            GlycanCompositionSpectrumSolutionSet.id, ondelete='CASCADE'),
        index=True)
    solution_set = relationship(GlycanCompositionSpectrumSolutionSet,
                                backref=backref("spectrum_matches", lazy='subquery'))

    composition_id = Column(
        Integer, ForeignKey(GlycanComposition.id, ondelete='CASCADE'),
        index=True)

    composition = relationship(GlycanComposition)

    @property
    def target(self):
        return self.composition

    @classmethod
    def serialize(cls, obj, session, scan_look_up_cache, mass_shift_cache, analysis_id,
                  solution_set_id, is_decoy=False, *args, **kwargs):
        inst = cls(
            scan_id=scan_look_up_cache[obj.scan.id],
            analysis_id=analysis_id,
            score=obj.score,
            solution_set_id=solution_set_id,
            composition_id=obj.target.id,
            mass_shift_id=mass_shift_cache[obj.mass_shift].id)
        session.add(inst)
        session.flush()
        return inst

    def convert(self, mass_shift_cache=None):
        session = object_session(self)
        scan = session.query(MSScan).get(self.scan_id).convert(False, False)
        mass_shift = self._resolve_mass_shift(session, mass_shift_cache)
        target = session.query(GlycanComposition).get(self.composition_id).convert()
        inst = MemorySpectrumMatch(scan, target, self.score, mass_shift=mass_shift)
        inst.id = self.id
        return inst

    def __repr__(self):
        return "DB" + repr(self.convert())


class UnidentifiedSpectrumCluster(Base, SpectrumClusterBase, BoundToAnalysis):
    __tablename__ = "UnidentifiedSpectrumCluster"

    id = Column(Integer, primary_key=True)

    @classmethod
    def serialize(cls, obj, session, scan_look_up_cache, mass_shift_cache, analysis_id, *args, **kwargs):
        inst = cls()
        session.add(inst)
        session.flush()
        cluster_id = inst.id
        for solution_set in obj.tandem_solutions:
            UnidentifiedSpectrumSolutionSet.serialize(
                solution_set, session, scan_look_up_cache, mass_shift_cache, analysis_id,
                cluster_id, *args, **kwargs)
        return inst

    source = relationship(
        "UnidentifiedChromatogram",
        secondary=lambda: UnidentifiedChromatogramToUnidentifiedSpectrumCluster,
        backref=backref("spectrum_cluster", uselist=False))


class UnidentifiedSpectrumSolutionSet(Base, SolutionSetBase, BoundToAnalysis):
    __tablename__ = "UnidentifiedSpectrumSolutionSet"

    id = Column(Integer, primary_key=True)
    cluster_id = Column(
        Integer,
        ForeignKey(UnidentifiedSpectrumCluster.id, ondelete="CASCADE"),
        index=True)

    cluster = relationship(UnidentifiedSpectrumCluster, backref=backref(
        "spectrum_solutions", lazy='subquery'))

    @classmethod
    def serialize(cls, obj, session, scan_look_up_cache, mass_shift_cache, analysis_id, cluster_id, *args, **kwargs):
        inst = cls(
            scan_id=scan_look_up_cache[obj.scan.id],
            analysis_id=analysis_id,
            cluster_id=cluster_id)
        session.add(inst)
        session.flush()
        # if we have a real SpectrumSolutionSet, then it will be iterable
        try:
            list(obj)
        except TypeError:
            # otherwise we have a single SpectrumMatch
            obj = [obj]
        for solution in obj:
            UnidentifiedSpectrumMatch.serialize(
                solution, session, scan_look_up_cache, mass_shift_cache,
                analysis_id, inst.id, *args, **kwargs)
        return inst

    def convert(self, mass_shift_cache=None, scan_cache=None, structure_cache=None):
        if scan_cache is None:
            scan_cache = dict()
        if mass_shift_cache is None:
            mass_shift_cache = dict()
        matches = [x.convert(mass_shift_cache=mass_shift_cache,
                             scan_cache=scan_cache) for x in self.spectrum_matches]
        matches.sort(key=lambda x: x.score, reverse=True)
        inst = MemorySpectrumSolutionSet(
            convert_scan_to_record(self.scan),
            matches
        )
        inst.q_value = min(x.q_value for x in inst)
        inst.id = self.id
        return inst

    def __repr__(self):
        return "DB" + repr(self.convert())


class UnidentifiedSpectrumMatch(Base, SpectrumMatchBase):
    __tablename__ = "UnidentifiedSpectrumMatch"

    id = Column(Integer, primary_key=True)
    solution_set_id = Column(
        Integer, ForeignKey(
            UnidentifiedSpectrumSolutionSet.id, ondelete='CASCADE'),
        index=True)

    solution_set = relationship(UnidentifiedSpectrumSolutionSet,
                                backref=backref("spectrum_matches", lazy='subquery'))

    @classmethod
    def serialize(cls, obj, session, scan_look_up_cache, mass_shift_cache, analysis_id, solution_set_id,
                  is_decoy=False, *args, **kwargs):
        inst = cls(
            scan_id=scan_look_up_cache[obj.scan.id],
            analysis_id=analysis_id,
            score=obj.score,
            solution_set_id=solution_set_id,
            mass_shift_id=mass_shift_cache[obj.mass_shift].id)
        session.add(inst)
        session.flush()
        return inst

    def convert(self, mass_shift_cache=None):
        session = object_session(self)
        scan = session.query(MSScan).get(self.scan_id).convert(False, False)
        mass_shift = self._resolve_mass_shift(session, mass_shift_cache)
        inst = MemorySpectrumMatch(scan, None, self.score, mass_shift=mass_shift)
        inst.id = self.id
        return inst

    def __repr__(self):
        return "DB" + repr(self.convert())


GlycanCompositionChromatogramToGlycanCompositionSpectrumCluster = Table(
    "GlycanCompositionChromatogramToGlycanCompositionSpectrumCluster", Base.metadata,
    Column("chromatogram_id", Integer, ForeignKey(
        "GlycanCompositionChromatogram.id", ondelete="CASCADE"), primary_key=True),
    Column("cluster_id", Integer, ForeignKey(
        GlycanCompositionSpectrumCluster.id, ondelete="CASCADE"), primary_key=True))


UnidentifiedChromatogramToUnidentifiedSpectrumCluster = Table(
    "UnidentifiedChromatogramToUnidentifiedSpectrumCluster", Base.metadata,
    Column("chromatogram_id", Integer, ForeignKey(
        "UnidentifiedChromatogram.id", ondelete="CASCADE"), primary_key=True),
    Column("cluster_id", Integer, ForeignKey(
        UnidentifiedSpectrumCluster.id, ondelete="CASCADE"), primary_key=True))


class GlycopeptideSpectrumMatchScoreSet(Base):
    __tablename__ = "GlycopeptideSpectrumMatchScoreSet"

    id = Column(Integer, ForeignKey(GlycopeptideSpectrumMatch.id), primary_key=True)
    peptide_score = Column(Numeric(12, 6, asdecimal=False), index=False)
    glycan_score = Column(Numeric(12, 6, asdecimal=False), index=False)
    glycopeptide_score = Column(Numeric(12, 6, asdecimal=False), index=False)
    glycan_coverage = needs_migration(Column(Numeric(8, 7, asdecimal=False), index=False), 0.0)

    total_q_value = Column(Numeric(8, 7, asdecimal=False), index=False)
    peptide_q_value = Column(Numeric(8, 7, asdecimal=False), index=False)
    glycan_q_value = Column(Numeric(8, 7, asdecimal=False), index=False)
    glycopeptide_q_value = Column(Numeric(8, 7, asdecimal=False), index=False)

    spectrum_match = relationship(
        GlycopeptideSpectrumMatch, backref=backref("score_set", lazy='joined', uselist=False))

    @classmethod
    def get_fields_from_object(cls, obj, db_id=None):
        scores = obj.score_set
        qs = obj.q_value_set
        fields = dict(
            id=db_id,
            peptide_score=scores.peptide_score,
            glycan_score=scores.glycan_score,
            glycan_coverage=scores.glycan_coverage,
            glycopeptide_score=scores.glycopeptide_score,
            total_q_value=qs.total_q_value,
            peptide_q_value=qs.peptide_q_value,
            glycan_q_value=qs.glycan_q_value,
            glycopeptide_q_value=qs.glycopeptide_q_value
        )
        return fields

    @classmethod
    def serialize_from_spectrum_match(cls, obj, session, db_id):
        fields = cls.get_fields_from_object(obj, db_id)
        session.execute(cls.__table__.insert(), fields)

    def convert(self):
        score_set = ScoreSet(
            peptide_score=self.peptide_score,
            glycan_score=self.glycan_score,
            glycopeptide_score=self.glycopeptide_score,
            glycan_coverage=self.glycan_coverage)
        fdr_set = FDRSet(
            total_q_value=self.total_q_value,
            peptide_q_value=self.peptide_q_value,
            glycan_q_value=self.glycan_q_value,
            glycopeptide_q_value=self.glycopeptide_q_value)
        return score_set, fdr_set

    def __repr__(self):
        return "DB" + repr(self.convert())

    @classmethod
    def bulk_load(cls, session, ids, chunk_size: int=512) -> Dict[int, Tuple[ScoreSet, FDRSet]]:
        out = {}
        for chunk in chunkiter(ids, chunk_size):
            res = session.execute(cls.__table__.select().where(cls.id.in_(chunk)))
            for self in res.fetchall():
                score_set = ScoreSet(
                    peptide_score=self.peptide_score,
                    glycan_score=self.glycan_score,
                    glycopeptide_score=self.glycopeptide_score,
                    glycan_coverage=self.glycan_coverage)
                fdr_set = FDRSet(
                    total_q_value=self.total_q_value,
                    peptide_q_value=self.peptide_q_value,
                    glycan_q_value=self.glycan_q_value,
                    glycopeptide_q_value=self.glycopeptide_q_value)
                out[self.id] = (score_set, fdr_set)
        return out


class LocalizationScore(Base):
    __tablename__ = "LocalizationScore"

    id = Column(Integer, ForeignKey(
        GlycopeptideSpectrumMatch.id), primary_key=True)

    position = Column(Integer, primary_key=True)
    modification = Column(String)
    score = Column(Numeric(6, 3, asdecimal=False))

    spectrum_match = relationship(
        GlycopeptideSpectrumMatch, backref=backref("localizations", lazy='joined'))

    @classmethod
    def get_fields_from_object(cls, obj: MemorySpectrumMatch, db_id=None):
        instances = []
        for loc in obj.localizations:
            inst = dict(
                id=db_id,
                position=loc.position,
                modification=loc.modification,
                score=loc.score
            )
            instances.append(inst)
        return instances

    def convert(self):
        return MemoryLocalizationScore(self.position, self.modification, self.score)

    def __repr__(self):
        return f"DB{self.__class__.__name__}({self.id}, {self.position}, {self.modification}, {self.score})"

    def __str__(self):
        return f"{self.modification}:{self.position}:{self.score}"

    @classmethod
    def bulk_load(cls, session, ids, chunk_size: int = 512) -> DefaultDict[int, List[MemoryLocalizationScore]]:
        out = DefaultDict(list)
        for chunk in chunkiter(ids, chunk_size):
            res = session.execute(
                cls.__table__.select().where(cls.id.in_(chunk)))
            for self in res.fetchall():
                out[self.id].append(MemoryLocalizationScore(
                    self.position, self.modification, self.score))
        return out



import re

from collections import OrderedDict

from typing import Any

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import (
    relationship, backref, Query,
    deferred, object_session)
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method
from sqlalchemy import (
    Column, Numeric, Integer, String, ForeignKey, PickleType,
    Text, Index, select)
from sqlalchemy.ext.mutable import MutableDict, MutableList

from glycresoft.serialize.base import (
    Base)


from .hypothesis import GlycopeptideHypothesis
from .glycan import GlycanCombination
from .generic import JSONType, HasChemicalComposition
from ..utils import chunkiter

from glycopeptidepy.structure import sequence, residue, PeptideSequenceBase
from glycopeptidepy.io.annotation import AnnotationCollection, GlycosylationSite, GlycosylationType
from glycresoft.structure.structure_loader import PeptideProteinRelation, FragmentCachingGlycopeptide
from glycresoft.structure.utils import LRUDict


class AminoAcidSequenceWrapperBase(PeptideSequenceBase):
    _sequence_obj = None

    def _get_sequence_str(self):
        raise NotImplementedError()

    def _get_sequence(self):
        if self._sequence_obj is None:
            self._sequence_obj = sequence.PeptideSequence(
                self._get_sequence_str())
        return self._sequence_obj

    def __iter__(self):
        return iter(self._get_sequence())

    def __getitem__(self, i):
        return self._get_sequence()[i]

    def __len__(self):
        return len(self._get_sequence())

    def __str__(self):
        return str(self._get_sequence())


class Protein(Base, AminoAcidSequenceWrapperBase):
    __tablename__ = "Protein"

    id = Column(Integer, primary_key=True, autoincrement=True)
    protein_sequence = Column(Text, default=u"")
    name = Column(String(128), index=True)
    other = Column(MutableDict.as_mutable(PickleType))
    hypothesis_id = Column(Integer, ForeignKey(
        GlycopeptideHypothesis.id, ondelete="CASCADE"))
    hypothesis = relationship(GlycopeptideHypothesis, backref=backref('proteins', lazy='dynamic'))

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        annotations = kwargs.pop("annotations", AnnotationCollection([]))
        self.other = MutableDict()
        super().__init__(*args, **kwargs)
        self.annotations = annotations

    def _get_sequence_str(self):
        return self.protein_sequence.replace("Z", "X")

    _n_glycan_sequon_sites = None

    @property
    def n_glycan_sequon_sites(self):
        if self._n_glycan_sequon_sites is None:
            sites = self.sites.filter(ProteinSite.name == ProteinSite.N_GLYCOSYLATION).all()
            if sites:
                self._n_glycan_sequon_sites = sorted([int(i) for i in sites])
            # else:
            #     try:
            #         self._n_glycan_sequon_sites = sequence.find_n_glycosylation_sequons(self._get_sequence())
            #     except residue.UnknownAminoAcidException:
            #         self._n_glycan_sequon_sites = []
            else:
                self._n_glycan_sequon_sites = []
        return self._n_glycan_sequon_sites

    _o_glycan_sequon_sites = None

    @property
    def o_glycan_sequon_sites(self):
        if self._o_glycan_sequon_sites is None:
            sites = self.sites.filter(ProteinSite.name == ProteinSite.O_GLYCOSYLATION).all()
            if sites:
                self._o_glycan_sequon_sites = sorted([int(i) for i in sites])
            else:
                try:
                    self._o_glycan_sequon_sites = sequence.find_o_glycosylation_sequons(self._get_sequence())
                except residue.UnknownAminoAcidException:
                    self._o_glycan_sequon_sites = []
        return self._o_glycan_sequon_sites

    _glycosaminoglycan_sequon_sites = None

    @property
    def glycosaminoglycan_sequon_sites(self):
        if self._glycosaminoglycan_sequon_sites is None:
            sites = self.sites.filter(ProteinSite.name == ProteinSite.GAGYLATION).all()
            if sites:
                self._glycosaminoglycan_sequon_sites = sorted([int(i) for i in sites])
            else:
                try:
                    self._glycosaminoglycan_sequon_sites = sequence.find_glycosaminoglycan_sequons(
                        self._get_sequence())
                except residue.UnknownAminoAcidException:
                    self._glycosaminoglycan_sequon_sites = []
        return self._glycosaminoglycan_sequon_sites

    @property
    def glycosylation_sites(self):
        try:
            return self.n_glycan_sequon_sites  # + self.o_glycan_sequon_sites
        except residue.UnknownAminoAcidException:
            return []

    def _init_sites(self, include_cysteine_n_glycosylation: bool = False):
        try:
            parsed_sequence = self._get_sequence()
        except residue.UnknownAminoAcidException:
            return
        sites = []
        try:
            n_glycosites = sequence.find_n_glycosylation_sequons(
                parsed_sequence, include_cysteine=include_cysteine_n_glycosylation)
            for n_glycosite in n_glycosites:
                sites.append(
                    ProteinSite(name=ProteinSite.N_GLYCOSYLATION, location=n_glycosite))
        except residue.UnknownAminoAcidException:
            pass
        if self.annotations is not None:
            for modsite in self.annotations.modifications():
                if isinstance(modsite, GlycosylationSite):
                    if modsite.glycosylation_type == GlycosylationType.n_linked:
                        sites.append(
                            ProteinSite(name=ProteinSite.N_GLYCOSYLATION, location=modsite.position)
                        )
                    elif modsite.glycosylation_type == GlycosylationType.o_linked:
                        sites.append(
                            ProteinSite(name=ProteinSite.O_GLYCOSYLATION, location=modsite.position)
                        )
                    elif modsite.glycosylation_type == GlycosylationType.glycosaminoglycan:
                        sites.append(
                            ProteinSite(name=ProteinSite.GAGYLATION, location=modsite.position)
                        )

        # The O- and GAG-linker sites are not determined by a multi AA sequon. We don't
        # need to abstract them away and they are much too common.
        # try:
        #     o_glycosites = sequence.find_o_glycosylation_sequons(
        #         parsed_sequence)
        #     for o_glycosite in o_glycosites:
        #         sites.append(
        #             ProteinSite(name=ProteinSite.O_GLYCOSYLATION, location=o_glycosite))
        # except residue.UnknownAminoAcidException:
        #     pass

        # try:
        #     gag_sites = sequence.find_glycosaminoglycan_sequons(
        #         parsed_sequence)
        #     for gag_site in gag_sites:
        #         sites.append(
        #             ProteinSite(name=ProteinSite.GAGYLATION, location=gag_site))
        # except residue.UnknownAminoAcidException:
        #     pass
        self.sites.extend(sorted(set(sites)))

    def __repr__(self):
        return "DBProtein({0}, {1}, {2}, {3}...)".format(
            self.id, self.name, self.glycosylation_sites,
            self.protein_sequence[:20] if self.protein_sequence is not None else "")

    def to_json(self, full=False):
        d = OrderedDict((
            ('id', self.id),
            ('name', self.name),
            ("glycosylation_sites", list(self.glycosylation_sites)),
            ('other', self.other)
        ))
        if full:
            d.update({
                "protein_sequence": self.protein_sequence
            })
            for k, v in self.__dict__.items():
                if isinstance(v, Query):
                    d[k + '_count'] = v.count()
        return d

    def reverse(self, copy_id=False, prefix=None, suffix=None):
        n = len(self.protein_sequence)
        sites = []
        for site in self.sites:  # pylint: disable=access-member-before-definition
            sites.append(site.__class__(name=site.name, location=n - site.location - 1))
        name = self.name
        if name.startswith(">"):
            if prefix:
                name = ">" + prefix + name[1:]
        if suffix:
            name = name + suffix

        inst = self.__class__(name=name, protein_sequence=self.protein_sequence[::-1])
        if copy_id:
            inst.id = self.id
        inst.sites = sites
        return inst

    @property
    def annotations(self) -> AnnotationCollection:
        return self.other.get('annotations', AnnotationCollection([]))

    @annotations.setter
    def annotations(self, value):
        if not isinstance(value, AnnotationCollection) and value is not None:
            value = AnnotationCollection(value)
        elif value is None:
            value = AnnotationCollection([])
        self.other['annotations'] = value


class ProteinSite(Base):
    __tablename__ = "ProteinSite"

    id = Column(Integer, primary_key=True)
    name = Column(String(32), index=True)
    location = Column(Integer, index=True)
    protein_id = Column(Integer, ForeignKey(Protein.id, ondelete="CASCADE"), index=True)
    protein = relationship(Protein, backref=backref("sites", lazy='dynamic'))

    N_GLYCOSYLATION = "N-Glycosylation"
    O_GLYCOSYLATION = "O-Glycosylation"
    GAGYLATION = "Glycosaminoglycosylation"

    _hash = None

    def __repr__(self):
        return ("{self.__class__.__name__}(location={self.location}, "
                "name={self.name})").format(self=self)

    def __hash__(self):
        hash_val = self._hash
        if hash_val is None:
            hash_val = self._hash = hash((self.name, self.location))
        return hash_val

    def __index__(self):
        return self.location

    def __int__(self):
        return self.location

    def __add__(self, other):
        return int(self) + int(other)

    def __radd__(self, other):
        return int(self) + int(other)

    def __lt__(self, other):
        return int(self) < int(other)

    def __gt__(self, other):
        return int(self) > int(other)

    def __eq__(self, other):
        if isinstance(other, ProteinSite):
            return self.location == other.location and self.name == other.name
        return int(self) == int(other)

    def __ne__(self, other):
        return not (self == other)


def _convert_class_name_to_collection_name(name):
    parts = re.split(r"([A-Z]+[a-z]+)", name)
    parts = [p.lower() for p in parts if p]
    return '_'.join(parts) + 's'


class PeptideBase(AminoAcidSequenceWrapperBase):
    @declared_attr
    def protein_id(self):
        return Column(Integer, ForeignKey(
            Protein.id, ondelete="CASCADE"), index=True)

    @declared_attr
    def hypothesis_id(self):
        return Column(Integer, ForeignKey(
            GlycopeptideHypothesis.id, ondelete="CASCADE"), index=True)

    @declared_attr
    def protein(self):
        if not hasattr(self, "__collection_name__"):
            name = _convert_class_name_to_collection_name(self.__name__)
        else:
            name = self.__collection_name__
        return relationship(Protein, backref=backref(name, lazy='dynamic'))

    calculated_mass = Column(Numeric(12, 6, asdecimal=False), index=True)
    formula = Column(String(128))

    def __iter__(self):
        return iter(self.convert())

    def __len__(self):
        return len(self.convert())

    @property
    def total_mass(self):
        return self.convert().total_mass

    _protein_relation = None

    @property
    def protein_relation(self):
        if self._protein_relation is None:
            peptide = self
            self._protein_relation = PeptideProteinRelation(
                peptide.start_position, peptide.end_position, peptide.protein_id,
                peptide.hypothesis_id)
        return self._protein_relation

    @property
    def start_position(self):
        return self.protein_relation.start_position

    @property
    def end_position(self):
        return self.protein_relation.end_position

    def overlaps(self, other):
        return self.protein_relation.overlaps(other.protein_relation)

    def spans(self, position):
        return position in self.protein_relation


class Peptide(PeptideBase, HasChemicalComposition, Base):
    __tablename__ = 'Peptide'

    id = Column(Integer, primary_key=True)

    count_glycosylation_sites = Column(Integer)
    count_missed_cleavages = Column(Integer)
    count_variable_modifications = Column(Integer)

    start_position = Column(Integer)
    end_position = Column(Integer)

    peptide_score = Column(Numeric(12, 6, asdecimal=False))
    _scores = deferred(Column("scores", JSONType))
    peptide_score_type = Column(String(56))

    base_peptide_sequence = Column(String(512))
    modified_peptide_sequence = Column(String(512))

    sequence_length = Column(Integer)

    n_glycosylation_sites = Column(MutableList.as_mutable(JSONType))
    o_glycosylation_sites = Column(MutableList.as_mutable(JSONType))
    gagylation_sites = Column(MutableList.as_mutable(JSONType))

    hypothesis = relationship(GlycopeptideHypothesis, backref=backref('peptides', lazy='dynamic'))

    def _get_sequence_str(self):
        return self.modified_peptide_sequence

    def convert(self):
        inst = sequence.parse(self.modified_peptide_sequence)
        inst.id = self.id
        return inst

    def __repr__(self):
        return ("DBPeptideSequence({self.modified_peptide_sequence}, {self.n_glycosylation_sites},"
                " {self.start_position}, {self.end_position})").format(self=self)

    __table_args__ = (
        Index("ix_Peptide_mass_search_index", "hypothesis_id", "calculated_mass"),
        Index("ix_Peptide_coordinate_index", "id", "calculated_mass",
              "start_position", "end_position"),)

    @hybrid_method
    def spans(self, position):
        return position in self.protein_relation

    @spans.expression
    def spans(self, position):
        # Overhang at the end is for consistency with SpanningMixin, but this is not
        # ideal
        return (self.start_position <= position) & (position <= self.end_position)

    @hybrid_property
    def scores(self): # pylint: disable=method-hidden
        try:
            if self._scores is None:
                self.scores = []
            return self._scores
        except Exception:
            return []

    @scores.setter
    def scores(self, value):
        self._scores = value


class Glycopeptide(PeptideBase, Base):
    __tablename__ = "Glycopeptide"

    id = Column(Integer, primary_key=True)
    peptide_id = Column(Integer, ForeignKey(Peptide.id, ondelete='CASCADE'), index=True)
    glycan_combination_id = Column(Integer, ForeignKey(GlycanCombination.id, ondelete='CASCADE'), index=True)

    peptide = relationship(Peptide, backref=backref("glycopeptides", lazy='dynamic'))
    glycan_combination = relationship(GlycanCombination)

    glycopeptide_sequence = Column(String(1024))

    hypothesis = relationship(GlycopeptideHypothesis, backref=backref('glycopeptides', lazy='dynamic'))

    def _get_sequence_str(self):
        return self.glycopeptide_sequence

    @classmethod
    def bulk_load(cls, session, ids, chunk_size: int=512, peptide_relation_cache=None, structure_cache=None):
        if peptide_relation_cache is None:
            peptide_relation_cache = session.info.get("peptide_relation_cache")
            if peptide_relation_cache is None:
                peptide_relation_cache = session.info['peptide_relation_cache'] = LRUDict(
                    maxsize=1024)
        if structure_cache is None:
            structure_cache = {}
        out = structure_cache
        peptide_ids = []
        for chunk in chunkiter(ids, chunk_size):
            chunk = list(filter(lambda x: x not in structure_cache, chunk))
            res = session.execute(cls.__table__.select().where(
                cls.id.in_(chunk)
                )
            )

            for self in res.fetchall():
                inst = FragmentCachingGlycopeptide(self.glycopeptide_sequence)
                inst.id = self.id

                peptide_id = self.peptide_id
                out[self.id] = (inst, peptide_id)
                if peptide_id not in peptide_relation_cache:
                    peptide_ids.append(peptide_id)

        for chunk in chunkiter(peptide_ids, chunk_size):
            res = session.execute(Peptide.__table__.select().where(
                Peptide.id.in_(
                    list(filter(lambda x: x not in peptide_relation_cache,
                                chunk)))))

            for self in res.fetchall():
                peptide_relation_cache[self.id] = PeptideProteinRelation(
                    self.start_position,
                    self.end_position,
                    self.protein_id,
                    self.hypothesis_id
                )

        result = []
        for i in ids:
            inst, peptide_id = out[i]
            inst.protein_relation = peptide_relation_cache[peptide_id]
            result.append(inst)
        return result

    def convert(self, peptide_relation_cache=None):
        if peptide_relation_cache is None:
            session = object_session(self)
            peptide_relation_cache = session.info.get("peptide_relation_cache")
            if peptide_relation_cache is None:
                peptide_relation_cache = session.info['peptide_relation_cache'] = LRUDict(maxsize=1024)
        inst = FragmentCachingGlycopeptide(self.glycopeptide_sequence)
        inst.id = self.id
        peptide_id = self.peptide_id
        if peptide_id in peptide_relation_cache:
            inst.protein_relation = peptide_relation_cache[self.peptide_id]
        else:
            session = object_session(self)
            peptide_props = session.query(
                Peptide.start_position, Peptide.end_position,
                Peptide.protein_id, Peptide.hypothesis_id).filter(Peptide.id == peptide_id).first()
            peptide_relation_cache[peptide_id] = inst.protein_relation = PeptideProteinRelation(*peptide_props)
        return inst

    def __repr__(self):
        return "DBGlycopeptideSequence({self.glycopeptide_sequence}, {self.calculated_mass})".format(self=self)
    _protein_relation = None

    @hybrid_method
    def is_multiply_glycosylated(self):
        return self.glycan_combination.count > 1

    @is_multiply_glycosylated.expression
    def is_multiply_glycosylated(self):
        expr = select([GlycanCombination.count > 1]).where(
            GlycanCombination.id == Glycopeptide.glycan_combination_id).label(
            "is_multiply_glycosylated")
        return expr

    @property
    def protein_relation(self):
        if self._protein_relation is None:
            peptide = self.peptide
            self._protein_relation = PeptideProteinRelation(
                peptide.start_position, peptide.end_position, peptide.protein_id,
                peptide.hypothesis_id)
        return self._protein_relation

    @property
    def n_glycan_sequon_sites(self):
        return self.peptide.n_glycosylation_sites

    @property
    def o_glycan_sequon_sites(self):
        return self.peptide.o_glycosylation_sites

    @property
    def glycosaminoglycan_sequon_sites(self):
        return self.peptide.gagylation_sites

    @property
    def glycan_composition(self):
        return self.glycan_combination.convert()

    __table_args__ = (
        Index("ix_Glycopeptide_mass_search_index", "hypothesis_id", "calculated_mass"),
        # Index("ix_Glycopeptide_mass_search_index_full", "calculated_mass", "hypothesis_id",
        #                                                 "peptide_id", "glycan_combination_id"),
    )



import re
import threading
import multiprocessing

from queue import Empty
from typing import IO, List, Deque, Union

import urllib3
from ms_deisotope.data_source._compression import get_opener

from glycopeptidepy.io import uniprot, fasta, annotation
from glycopeptidepy.io.utils import UniprotToPeffConverter

from glycresoft.task import TaskBase
from glycresoft.serialize import Protein


uniprot_accession_pattern = re.compile(
    r"""(([OPQ]\d[A-Z0-9][A-Z0-9][A-Z0-9]\d)|
        ([A-Z0-9]\d[A-Z][A-Z0-9][A-Z0-9]\d)|
        ([A-NR-Z]\d[A-Z][A-Z0-9][A-Z0-9]\d[A-Z][A-Z0-9][A-Z0-9]\d))""",
        re.X)


def is_uniprot_accession(accession: str) -> bool:
    return bool(uniprot_accession_pattern.match(accession))


def get_uniprot_accession(name):
    try:
        peff_header = fasta.peff_parser(name)
        # Don't bother UniProt over PEFF specifications, they won't parse correctly
        # with the partial UniProt parser
        if peff_header:
            return None
    except fasta.UnparsableDeflineError:
        pass
    try:
        return fasta.partial_uniprot_parser(name).accession
    except (AttributeError, fasta.UnparsableDeflineError):
        # Some programs will just put the accession number into the mzIdentML
        # identifier, so we'll check if the first word looks like an accession
        # number and try it if so.
        accession = str(name).split(" ")[0]
        if is_uniprot_accession(accession):
            return accession
        return None


def retry(task, n=5):
    result = None
    errs = []
    for i in range(n):
        try:
            result = task()
            return result, errs
        except Exception as err:
            errs.append(err)
    raise errs[-1]


HAS_BATCH_HANDLER = hasattr(uniprot, 'get_features_for_many')


class UniprotSource(TaskBase):
    def task_handler(self, accession_number):
        accession = get_uniprot_accession(accession_number)
        if accession is None:
            accession = accession_number
        protein_data, errs = retry(lambda: uniprot.get(accession), 5)
        self.output_queue.put((accession_number, protein_data))
        return protein_data

    def batch_handler(self, accession_list):
        idents_map = {}
        for acc in accession_list:
            acc_ = get_uniprot_accession(acc)
            if acc_ is None:
                acc_ = acc
            idents_map[acc_] = acc
        idents = list(idents_map)
        result, errs = retry(lambda: uniprot.get_features_for_many(idents), 5)
        if errs:
            self.log(f"... Handled batch of size {len(accession_list)} and encountered {len(errs)} errors")

        for (acc, protein_data) in result:
            name = idents_map[acc]
            self.output_queue.put((name, protein_data))

    def fetch(self, accession_number):
        if isinstance(accession_number, list) and HAS_BATCH_HANDLER:
            try:
                self.batch_handler(accession_number)
            except Exception as e:
                self.error_handler(accession_number[0], e)
        else:
            try:
                self.task_handler(accession_number)
            except Exception as e:
                self.error_handler(accession_number, e)

    def error_handler(self, accession_number, error):
        self.output_queue.put((accession_number, Exception(str(error))))


class UniprotRequestingProcess(multiprocessing.Process, UniprotSource):
    process_name = "glycresoft-annotation-worker"

    def __init__(self, input_queue, output_queue, feeder_done_event, done_event):
        multiprocessing.Process.__init__(self)
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.feeder_done_event = feeder_done_event
        self.done_event = done_event

    def run(self):
        new_name = getattr(self, 'process_name', None)
        if new_name is not None:
            TaskBase().try_set_process_name(new_name)
        urllib3.disable_warnings()
        while True:
            try:
                accession = self.input_queue.get(True, 3)
            except Empty:
                if self.feeder_done_event.is_set():
                    break
            self.input_queue.task_done()
            self.fetch(accession)

        self.done_event.set()


def chunked(seq, size=128):
    n = len(seq)
    for i in range(0, n + size, size):
        z = seq[i:i + size]
        if z:
            yield z


class UniprotProteinDownloader(UniprotSource):
    accession_list: List[str]
    input_queue: multiprocessing.JoinableQueue
    output_queue: multiprocessing.Queue
    no_more_work: multiprocessing.Event
    done_event: multiprocessing.Event
    workers: List[UniprotRequestingProcess]

    def __init__(self, accession_list, n_threads=10):
        self.accession_list = accession_list
        self.n_threads = n_threads
        self.input_queue = multiprocessing.JoinableQueue(2000)
        self.output_queue = multiprocessing.Queue()
        self.no_more_work = multiprocessing.Event()
        self.done_event = multiprocessing.Event()
        self.workers = []

    def task_handler(self, accession_number):
        accession = get_uniprot_accession(accession_number)
        if accession is None:
            accession = accession_number
        protein_data = uniprot.get(accession)
        self.output_queue.put((accession_number, protein_data))
        return protein_data

    def fetch(self, accession_number):
        try:
            self.task_handler(accession_number)
        except Exception as e:
            self.error_handler(accession_number, e)

    def error_handler(self, accession_number, error):
        self.output_queue.put((accession_number, error))

    def feeder_task(self):
        n = len(self.accession_list)
        k = n // max(100, self.n_threads)
        k = max(min(k, 100), 1)
        self.log(f"... Submitting UniProt queries in batches of size {k}")

        for i, item in enumerate(chunked(self.accession_list, k)):
            if i % 1000 == 0 and i:
                self.input_queue.join()
            self.input_queue.put(item)
        self.no_more_work.set()

    def run(self):
        feeder = threading.Thread(target=self.feeder_task)
        feeder.start()
        self.workers = workers = []
        n = min(self.n_threads, len(self.accession_list))
        for i in range(n):
            t = UniprotRequestingProcess(
                self.input_queue, self.output_queue, self.no_more_work,
                multiprocessing.Event())
            t.daemon = True
            t.start()
            workers.append(t)

    def join(self):
        for worker in self.workers:
            if not worker.done_event.is_set():
                break
            else:
                worker.join()
        else:
            self.done_event.set()
            return True
        return False

    def start(self):
        self.run()

    def get(self, blocking=True, timeout=3):
        return self.output_queue.get(blocking, timeout)


class UniprotProteinXML(UniprotSource):
    path: Union[str, IO]
    store: annotation.AnnotationDatabase
    queue: Deque[str]
    done_event: threading.Event

    def __init__(self, path: str, ids: List[str]):
        self.path = path
        self.store = {}
        self.work = Deque(ids)
        self.load()
        self.done_event = threading.Event()

    def join(self):
        return

    def load(self):
        if self.path == '-':
            self.store = annotation.AnnotationDatabase({})
        else:
            stream = get_opener(self.path)
            if uniprot.is_uniprot_xml(stream):
                self.store = annotation.AnnotationDatabase.from_uniprot_xml(stream)
            else:
                self.store = annotation.AnnotationDatabase.load(stream)

    def fetch(self, accession_number: str):
        accession = get_uniprot_accession(accession_number)
        try:
            features = self.store[accession]
            return accession_number, features
        except KeyError:
            return accession_number, Exception(f"{accession_number!r} not found")

    def get(self, *args, **kwargs):
        if self.work:
            acc = self.work.popleft()
            if not self.work:
                self.done_event.set()
            return self.fetch(acc)
        else:
            raise Empty()


class UniprotProteinSource(TaskBase):
    def __init__(self, accession_list, hypothesis_id, n_threads=4):
        self._accession_list = accession_list
        self.n_threads = n_threads
        self.downloader = None
        self.hypothesis_id = hypothesis_id
        self.proteins = []

        self._clean_accession_list()

    def _clean_accession(self, accession_string):
        if accession_string.startswith(">"):
            accession_string = accession_string[1:]
        accession_string = accession_string.strip()
        result = None
        if "|" in accession_string:
            try:
                fields = fasta.default_parser(accession_string)
                result = fields.accession
            except fasta.UnparsableDeflineError:
                result = accession_string
        else:
            result = accession_string
        return result

    def _clean_accession_list(self):
        cleaned = []
        for entry in self._accession_list:
            cleaned.append(self._clean_accession(entry))
        self._accession_list = cleaned

    def _make_downloader(self):
        self.downloader = UniprotProteinDownloader(
            self._accession_list, self.n_threads)
        return self.downloader

    def make_protein(self, accession, uniprot_protein):
        sequence = uniprot_protein.sequence
        name = "sp|{accession}|{gene_name} {description}".format(
            accession=accession, gene_name=uniprot_protein.gene_name,
            description=uniprot_protein.recommended_name)
        protein = Protein(
            name=name, protein_sequence=sequence,
            hypothesis_id=self.hypothesis_id)
        return protein

    def run(self):
        downloader = self._make_downloader()
        downloader.start()

        has_work = True

        while has_work:
            try:
                accession, value = downloader.get(True, 3)
                if isinstance(value, Exception):
                    self.log("Could not retrieve %s - %r" % (accession, value))
                else:
                    protein = self.make_protein(accession, value)
                    self.proteins.append(protein)
            except Empty:
                # If we haven't completed the download process, block
                # now and wait for the threads to join, then continue
                # trying to fetch results
                if not downloader.done_event.is_set():
                    downloader.join()
                    continue
                # Otherwise we've already waited for all the results to
                # arrive and we can stop iterating
                else:
                    has_work = False


class UniprotToPEFFTranslator(TaskBase):
    input_fasta_path: str
    output_path: str
    n_processes: int

    def __init__(self, input_fasta_path, output_path, n_processes: int=12) -> None:
        super().__init__()
        self.input_fasta_path = input_fasta_path
        self.output_path = output_path
        self.n_processes = n_processes

    def run(self):
        reader = fasta.ProteinFastaFileReader(self.input_fasta_path, index=True)

        accessions = []
        for key in reader.index.keys():
            acc = key.get("accession")
            if acc:
                accessions.append(acc)

        queue = UniprotProteinDownloader(accessions, self.n_processes)

        queue.start()

        header_block = fasta.PEFFHeaderBlock()
        header_block['Prefix'] = 'sp'
        header_block['SequenceType'] = 'AA'

        converter = UniprotToPeffConverter()

        with open(self.output_path, 'wb') as fh:
            writer = fasta.PEFFWriter(fh)
            writer.write_header([header_block])
            while True:
                try:
                    name, rec = queue.get()
                    try:
                        prot = converter(rec)
                        writer.write(prot)
                    except fasta.UnknownAminoAcidException as err:
                        self.log(f"Skipping {name}, failed to convert to PEFF: {err}")

                except Empty:
                    queue.join()
                    if queue.done_event.is_set():
                        break
